from __future__ import annotations

from ctypes import POINTER, cast
from typing import TypeVar

from einspect.structs import PyObject
from einspect.structs.py_set import PySetObject, SetEntry
from einspect.types import Array, ptr
from einspect.views.unsafe import unsafe
from einspect.views.view_base import REF_DEFAULT, View

__all__ = ("SetView",)

_T = TypeVar("_T")


class SetView(View[set, None, _T]):
    _pyobject: PySetObject[_T]

    def __init__(self, obj: set[_T], ref: bool = REF_DEFAULT) -> None:
        super().__init__(obj, ref)

    def __len__(self) -> int:
        return self.used

    @property
    def fill(self) -> int:
        """Number active and dummy entries."""
        return self._pyobject.fill

    @fill.setter
    @unsafe
    def fill(self, value: int) -> None:
        """Number active and dummy entries."""
        self._pyobject.fill = value

    @property
    def used(self) -> int:
        return self._pyobject.used

    @used.setter
    @unsafe
    def used(self, value: int) -> None:
        self._pyobject.used = value

    @property
    def mask(self) -> int:
        return self._pyobject.mask

    @mask.setter
    @unsafe
    def mask(self, value: int) -> None:
        self._pyobject.mask = value

    @property
    def table(self) -> Array[SetEntry[_T]]:
        size = self.mask + 1
        arr = cast(self._pyobject.table, POINTER(SetEntry * size))
        return arr.contents

    @table.setter
    @unsafe
    def table(self, value: Array[SetEntry[_T]]) -> None:
        self._pyobject.table = ptr(value[0])

    @property
    def finger(self) -> int:
        """Search finger for pop()"""
        return self._pyobject.finger

    @finger.setter
    @unsafe
    def finger(self, value: int) -> None:
        """Set search finger for pop()"""
        self._pyobject.finger = value

    @property
    def smalltable(self) -> Array[SetEntry[_T]]:
        return self._pyobject.smalltable

    @smalltable.setter
    @unsafe
    def smalltable(self, value: Array[SetEntry[_T]]) -> None:
        self._pyobject.smalltable = value

    @property
    def weakreflist(self) -> ptr[PyObject]:
        return self._pyobject.weakreflist

    @weakreflist.setter
    @unsafe
    def weakreflist(self, value: ptr[PyObject]) -> None:
        self._pyobject.weakreflist = value



from __future__ import annotations

from ctypes import PYFUNCTYPE, Structure, c_char_p, c_int, py_object

from typing_extensions import Annotated

from einspect.structs import struct

__all__ = ("PyMethodDef", "PyCFunction")

PyCFunction = PYFUNCTYPE(py_object, py_object, py_object)


@struct
class PyMethodDef(Structure):
    ml_name: Annotated[bytes, c_char_p]
    ml_meth: PyCFunction
    ml_flags: Annotated[int, c_int]
    ml_doc: Annotated[bytes, c_char_p]



from __future__ import annotations

from einspect.structs import PyLongObject
from einspect.structs.deco import struct

__all__ = ("PyBoolObject",)


@struct
class PyBoolObject(PyLongObject):
    """
    Defines a PyBoolObject Structure.

    https://github.com/python/cpython/blob/3.11/Objects/boolobject.c#L198-L206
    """



from __future__ import annotations

from collections.abc import Sequence
from ctypes import POINTER, addressof, memmove, sizeof
from typing import TypeVar, overload

from einspect.api import seq_to_array
from einspect.compat import abc
from einspect.structs import PyListObject, PyObject
from einspect.types import Array, ptr
from einspect.views.unsafe import unsafe
from einspect.views.view_base import REF_DEFAULT, VarView

__all__ = ("ListView",)

_VT = TypeVar("_VT")


class ListView(VarView[list, None, _VT], abc.Sequence[_VT]):
    _pyobject: PyListObject[_VT]

    def __init__(self, obj: list[_VT], ref: bool = REF_DEFAULT) -> None:
        super().__init__(obj, ref)

    @overload
    def __getitem__(self, index: int) -> _VT:
        ...

    @overload
    def __getitem__(self, index: slice) -> list[_VT]:
        ...

    def __getitem__(self, index: int | slice) -> _VT | list[_VT]:
        if isinstance(index, int):
            obj_ptr = self._pyobject.GetItem(index)
            return obj_ptr.contents.into_object()
        elif isinstance(index, slice):
            # Normalize slice start and stop
            start = index.start if index.start is not None else 0
            stop = index.stop if index.stop is not None else self.size
            obj_ptr = self._pyobject.GetSlice(start, stop)
            ls = obj_ptr.contents.into_object()
            # Get step if provided
            if index.step is not None:
                ls = ls[:: index.step]
            return ls
        raise TypeError(f"Invalid index type: {type(index)}")

    def __setitem__(self, index: int, value: _VT) -> None:
        if isinstance(index, int):
            self._pyobject.SetItem(index.__index__(), value)
        elif isinstance(index, slice):
            if index.step is not None:
                raise ValueError("Cannot set slice with step")
            # Normalize slice start and stop
            start = index.start if index.start is not None else 0
            stop = index.stop if index.stop is not None else self.size
            self._pyobject.SetSlice(start, stop, value)
        else:
            raise TypeError(f"Invalid index type: {type(index)}")

    def __len__(self) -> int:
        return self.size

    @property
    def allocated(self) -> int:
        """Allocated size of the list."""
        return self._pyobject.allocated  # type: ignore

    @allocated.setter
    @unsafe
    def allocated(self, value: int) -> None:
        """Set the allocated size of the list."""
        self._pyobject.allocated = value

    @property
    def item(self) -> Array[ptr[PyObject[_VT]]]:
        """Array of item pointers in the list."""
        arr_type = POINTER(PyObject) * self.size
        arr = arr_type.from_address(addressof(self._pyobject.ob_item.contents))
        return arr

    @item.setter
    @unsafe
    def item(
        self, value: Array[ptr[PyObject]] | Sequence[ptr[PyObject]] | Sequence[object]
    ) -> None:
        if not isinstance(value, Array):
            value = [PyObject.try_from(v).with_ref().as_ref() for v in value]
            value = seq_to_array(value, POINTER(PyObject))
        memmove(self._pyobject.ob_item, value, sizeof(value))



from __future__ import annotations

import logging
from ctypes import POINTER, Structure
from functools import partial
from typing import Callable, Literal, Sequence, Tuple, Type, TypeVar, Union, overload

# noinspection PyUnresolvedReferences, PyProtectedMember
from typing_extensions import _AnnotatedAlias, get_args, get_type_hints

from einspect.protocols.type_parse import convert_type_hints, fix_ctypes_generics
from einspect.types import _SelfPtr

__all__ = ("struct",)

log = logging.getLogger(__name__)

_T = TypeVar("_T", bound=Type[Structure])

FieldsType = Sequence[Union[Tuple[str, type], Tuple[str, type, int]]]


@overload
def struct(*, fields: FieldsType) -> Callable[[_T], _T]:
    ...


@overload
def struct(cls: _T, fields: Literal[None] = ...) -> _T:
    ...


def struct(cls: _T | None = None, fields: FieldsType | None = None):
    """Decorator to declare _fields_ on Structures via type hints."""
    # Normal decorator usage
    if cls is not None:
        return _struct(cls)

    # Usage as a decorator factory
    return partial(_struct, __fields=fields)


def _struct(cls: _T, __fields: FieldsType | None = None) -> _T:
    """Decorator to declare _fields_ on Structures via type hints."""
    # if fields provided, replace type hints with temp sentinel
    fields_overrides = {}
    for tup in __fields or ():
        f_name, f_type, *_ = tup
        f_ls = list(tup)
        # Convert SelfPtr types to POINTER(cls)
        if f_type is _SelfPtr:
            f_ls[1] = POINTER(cls)  # type: ignore

        fields_overrides[f_name] = tuple(f_ls)

    if __fields is not None:
        for tup in __fields:
            # This is to prevent errors during get_type_hints if there is an override
            cls.__annotations__[tup[0]] = None

    fields = []
    # Locals dict for type hint resolution
    hint_locals = {cls.__name__: cls}
    try:
        hints = get_type_hints(cls, None, hint_locals, include_extras=True)
    except (TypeError, NameError):
        # Normalize annotations of py_object subscripts
        fix_ctypes_generics(cls.__annotations__, cls.__name__)
        hints = get_type_hints(cls, None, hint_locals, include_extras=True)

    for name, type_hint in hints.items():
        # Use override if exists
        if f := fields_overrides.get(name):
            fields.append(f)
            continue
        # Skip actual values like _fields_
        if name.startswith("_") and name.endswith("_"):
            continue
        # Skip callables
        if type_hint == Callable:
            continue
        # Since get_type_hints also gets superclass hints, skip them
        if name not in cls.__annotations__:
            continue

        # For Annotated, directly use fields 1 and 2
        if type(type_hint) is _AnnotatedAlias:
            args = get_args(type_hint)
            res = (name, *args[1:3])
            log.debug(f"Annotated: {type_hint} -> {res}")
            fields.append((name, *args[1:3]))
            continue

        type_hint = convert_type_hints(type_hint, cls)
        fields.append((name, type_hint))

    # We must only set this once as _fields_ is final
    try:
        cls._fields_ = fields
    except (TypeError, AttributeError) as err:
        raise TypeError(
            f"Failed to set _fields_ ({fields}) on {cls.__name__} -> {err}"
        ) from err

    return cls



from __future__ import annotations

import ctypes
import logging
import warnings
import weakref
from abc import ABC
from ctypes import py_object
from functools import cached_property
from typing import (
    Any,
    Final,
    Generic,
    Type,
    TypeVar,
    get_args,
    get_type_hints,
    overload,
)

from einspect.api import PTR_SIZE, Py, PyObj_FromPtr, align_size
from einspect.errors import DroppedReference, MovedError, UnsafeError
from einspect.structs import PyObject, PyTypeObject, PyVarObject, TpFlags
from einspect.views._display import Formatter
from einspect.views._moves import check_move, move
from einspect.views.unsafe import UnsafeContext, unsafe

__all__ = ("View", "VarView", "AnyView", "REF_DEFAULT")

log = logging.getLogger(__name__)

REF_DEFAULT: Final[bool] = True

_T = TypeVar("_T")
_KT = TypeVar("_KT")
_VT = TypeVar("_VT")

# For moves
_View = TypeVar("_View", bound="View")
_Obj = TypeVar("_Obj", bound=object)


def wrap_py_object(obj: _T | py_object[_T]) -> py_object[_T]:
    """Wrap non-py_object objects in a py_object."""
    if isinstance(obj, py_object):
        return obj
    return py_object(obj)


class BaseView(ABC, Generic[_T, _KT, _VT], UnsafeContext):
    """Base class for all views."""

    def __init__(self, obj: _T, ref: bool = REF_DEFAULT) -> None:
        super().__init__()
        # Stores base info for repr and errors
        self._base_type: type[_T] = type(obj)
        self._base_id = id(obj)
        # Get a reference if ref=True
        self._base: py_object[_T] | None
        self._base = wrap_py_object(obj) if ref else None
        # Attempt to get a weakref
        try:
            self._base_weakref = weakref.ref(obj)
        except TypeError:
            self._base_weakref = None


class View(BaseView[_T, _KT, _VT]):
    """
    View for Python objects.

    Notes:
        The _pyobject class annotation is used to determine
        the type of the underlying PyObject struct.
    """

    _pyobject: PyObject[_T, None, None]

    def __init__(self, obj: _T, ref: bool = REF_DEFAULT) -> None:
        super().__init__(obj, ref)

        struct_type = get_type_hints(self.__class__)["_pyobject"]
        # For unions, use first type
        # There's no easy way to check this in <3.9 without types.UnionType
        # So we just check if not a PyObject / generic alias
        if not getattr(struct_type, "from_object", None):
            struct_type = get_args(struct_type)[0]

        self._pyobject = struct_type.from_object(obj)
        self.__dropped = False
        _ = self.mem_allocated  # cache allocated property

    def __repr__(self) -> str:
        """Return a string representation of the view."""
        addr = self._pyobject.address
        py_obj_cls = self._pyobject.__class__.__name__
        # If we have an instance dict (subclass), include base type in repr
        has_dict = self._pyobject.ob_type.contents.tp_dictoffset != 0
        # Or if we are the `View` class
        base = ""
        if has_dict or self.__class__ is View:
            base = f"[{self._base_type.__name__}]"
        return f"{self.__class__.__name__}{base}(<{py_obj_cls} at {addr:#04x}>)"

    def info(self, types: bool = True, arr_max: int | None = 64) -> str:
        """
        Return a formatted info string of the view struct.

        Args:
            types: If True, include types as annotations.
            arr_max: Maximum length of Array elements to display.

        Returns:
            A formatted info string of the view struct.
        """
        return Formatter(types, arr_max).format_view(self)

    @property
    def ref_count(self) -> int:
        """Reference count of the object."""
        return int(self._pyobject.ob_refcnt)  # type: ignore

    @ref_count.setter
    @unsafe
    def ref_count(self, value: int) -> None:
        self._pyobject.ob_refcnt = value

    @property
    def type(self) -> Type[_T]:
        """Type of the object."""
        return self._pyobject.ob_type.contents.into_object()

    @type.setter
    @unsafe
    def type(self, value: Type) -> None:
        self._pyobject.ob_type = PyTypeObject.from_object(value).as_ref()

    @property
    def base(self) -> _T:
        """
        Returns the base object of the view.

        Requires either the View to be created with (ref=True)
        or the object to support weakrefs.

        Notes:
            If neither ref nor weakref are available, and called within an
            unsafe context, returns an object via memory address cast.
            The result of the cast is undefined behavior, and could cause
            a segmentation fault.

        Returns:
            The base object of the view.

        Raises:
            AttributeError: If ref=False and base does not support weakrefs.
            MovedError: If weak-ref of base is garbage collected.
        """
        # Prioritize strong ref if it exists
        if self._base is not None:
            return self._base.value
        # If no weakref, error if no unsafe context
        if self._base_weakref is None:
            if not self._unsafe:
                raise UnsafeError(
                    f"Base object {self._base_type.__name__!r} does not support weak-refs, "
                    "use (ref=True) or an unsafe context to access base via memory address."
                ) from None
            else:
                # Give a resource warning if ref_count is <= 0
                ref_count = self.ref_count
                if ref_count <= 0:
                    warnings.warn(
                        f"Base object {self._base_type.__name__!r} has ref_count <= 0, "
                        "Accessing base via memory address is undefined behavior.",
                        RuntimeWarning,
                    )
        # Attempt to use weakref if alive
        else:
            base = self._base_weakref()
            if base is not None:
                return base
            else:
                if not self._unsafe:
                    raise MovedError(
                        f"Weak-referenced base object {self._base_type.__name__!r} has been garbage collected. "
                        "use unsafe context to access base via memory address."
                    ) from None
                else:
                    # Resource warning
                    warnings.warn(
                        f"Weak-referenced base object {self._base_type.__name__!r} has been garbage collected. "
                        "Accessing base via memory address is undefined behavior.",
                        RuntimeWarning,
                    )

        return self._pyobject.into_object()

    @property
    def mem_size(self) -> int:
        """Memory size of the object in bytes."""
        return self._pyobject.mem_size

    @cached_property
    def mem_allocated(self) -> int:
        """Memory allocated for the object in bytes."""
        return align_size(self.mem_size)

    @property
    def instance_dict(self) -> dict[str, Any] | None:
        """Instance Dictionary of the object."""
        if self._pyobject.ob_type.contents.tp_dictoffset == 0:
            cls = self._base_type.__name__
            raise TypeError(
                f"Object type {cls!r} does not support an instance dictionary."
            )
        ref = self._pyobject.instance_dict()
        return ref.contents.into_object() if ref else None

    @instance_dict.setter
    def instance_dict(self, value: dict[str, Any]) -> None:
        """Set the instance dictionary of the object."""
        if self._pyobject.ob_type.contents.tp_dictoffset == 0:
            cls = self._base_type.__name__
            raise TypeError(
                f"Object type {cls!r} does not support an instance dictionary."
            )
        self._pyobject.instance_dict().contents = PyObject.from_object(value).with_ref()

    def is_gc(self) -> bool:
        """
        Returns True if the object implements the Garbage Collector protocol.

        If True, a PyGC_HEAD struct will precede the object struct in memory.
        """
        return self._pyobject.is_gc()

    def gc_may_be_tracked(self) -> bool:
        """
        Return True if the object may be tracked by
        the GC in the future, or already is.
        """
        return self._pyobject.gc_may_be_tracked()

    def gc_is_tracked(self) -> bool:
        """Returns True if the object is tracked by the GC."""
        return self._pyobject.is_gc() and self._pyobject.gc_is_tracked()

    def drop(self) -> None:
        """
        Drop all references to the base object.

        Notes:
            This is useful for when you want to drop the reference
            to the base object, but still want to access the view.
        """
        ref = DroppedReference(type(self))
        self._pyobject = ref  # type: ignore
        self._base = ref  # type: ignore
        self._base_weakref = None
        self.__dropped = True

    @unsafe
    def move_to(self, dst: View, start: int = PTR_SIZE) -> None:
        """
        Copy the object to another view's location.

        Args:
            dst: The destination view.
            start: The start offset in bytes to copy from.
                The default is pointer size to skip `ob_refcnt`
        """
        if not isinstance(dst, View):
            raise TypeError(f"Expected View, got {type(dst).__name__!r}")

        move(src=self._pyobject, dst=dst._pyobject, offset=start)

    @overload
    def move_from(self, other: _View) -> _View:
        ...

    @overload
    def move_from(self, other: _Obj) -> View[_Obj]:
        ...

    def move_from(self, other):
        """Moves data at other Viewable to this View."""
        from einspect.views import factory

        if not isinstance(other, View):
            other = factory.view(other)  # type: ignore

        # Check move safety if not in unsafe context
        if not self._unsafe:
            check_move(self, other)

        # Store our address
        addr = self._pyobject.address
        # Move other to our pyobject address
        with other.unsafe():
            other.move_to(self)
        # Increment other refcount
        other._pyobject.IncRef()
        # Return a new view of ourselves
        obj = PyObj_FromPtr(addr)
        Py.IncRef(obj)
        v = factory.view(obj)
        # Drop old view
        self.drop()
        return v

    @overload
    def swap(self, other: _View) -> _View:
        ...

    @overload
    def swap(self, other: _Obj) -> View[_Obj]:
        ...

    def swap(self, other):
        """Swaps data at other Viewable with this View."""
        from einspect.views import factory

        if not isinstance(other, View):
            other = factory.view(other)  # type: ignore

        # Save the other PyObject type
        other_py_type = type(other._pyobject)
        other_is_managed_dict = (
            other._pyobject.ob_type.contents.tp_flags & TpFlags.MANAGED_DICT
        )

        # Check safety of both moves
        if not self._unsafe:
            check_move(self, other)
            check_move(other, self)

        other._pyobject.IncRef()

        # Save the other view in a buffer
        buf = ctypes.create_string_buffer(other.mem_allocated)
        ctypes.memmove(buf, other._pyobject.address, other.mem_size)
        # Also save the other instance dict, if it exists
        other_dict = None
        if (other_dict_ptr := other._pyobject.instance_dict()) is not None:
            other_dict = other_dict_ptr.contents
            # IncRef so the dict stays alive
            other_dict.IncRef()

        # Move self to other
        move(src=self._pyobject, dst=other._pyobject)

        # Get a PyObject from buffer
        other_obj: PyObject = other_py_type.from_address(ctypes.addressof(buf))  # type: ignore

        # Move buffer to self (exclude instance dict, handle it later)
        move(src=other_obj, dst=self._pyobject, inst_dict=False)

        # Return a new view of ourselves
        new_view = factory.view(self._pyobject.into_object())

        # Restore other instance dict, if it exists
        if other_dict is not None:
            # For managed dict, use SetAttr
            if other_is_managed_dict:
                new_view._pyobject.SetAttr("__dict__", other_dict.into_object())
            # Otherwise use the instance dict pointer
            else:
                new_view._pyobject.instance_dict().contents = other_dict

        # Drop old view
        self.drop()
        return new_view

    @overload
    def __lshift__(self, other: _View) -> _View:
        ...

    @overload
    def __lshift__(self, other: _Obj) -> View[_Obj]:
        ...

    def __lshift__(self, other):
        """Moves data at other View to this View."""
        return self.move_from(other)

    def __invert__(self) -> _T:
        """Returns the base of this view as object."""
        # Prioritize strong ref if it exists
        if self._base is not None:
            return self._base.value
        return self.base


class VarView(View[_T, _KT, _VT]):
    _pyobject: PyVarObject[_T, _KT, _VT]

    @property
    def size(self) -> int:
        """Size (ob_size) of the PyVarObject."""
        return self._pyobject.ob_size

    @size.setter
    @unsafe
    def size(self, value: int) -> None:
        self._pyobject.ob_size = value


class AnyView(View[_T, None, None]):
    _pyobject: PyObject[_T, _KT, _VT]

    @property
    def mem_size(self) -> int:
        """
        Memory size of the object in bytes.

        Notes:
            This will require casting into a py_object to use __sizeof__.
            If (ref=False), and the object does not support weakrefs,
            accessing this attribute will require an unsafe context.
        """
        return object.__sizeof__(self.base)

    def __repr__(self) -> str:
        addr = self._pyobject.address
        py_obj_cls = self._pyobject.__class__.__name__
        base = f"[{self._base_type.__name__}]"
        return f"{self.__class__.__name__}{base}(<{py_obj_cls} at {addr:#04x}>)"



"""Runtime patches."""
from __future__ import annotations

import ctypes
from ctypes import POINTER, addressof, c_void_p, memmove, sizeof

from einspect import types
from einspect.structs.py_object import PyObject


class _Null_LP_PyObject(POINTER(PyObject)):
    _type_ = PyObject

    def __repr__(self) -> str:
        """Returns the string representation of the pointer."""
        return f"<NULL ptr[PyObject] at {addressof(self):#04x}>"

    def __eq__(self, other) -> bool:
        """Returns equal to other null pointers."""
        # noinspection PyUnresolvedReferences
        if not isinstance(other, ctypes._Pointer):
            return NotImplemented
        return not other


def run() -> None:
    new_null = _Null_LP_PyObject()
    PyObject.from_object(new_null).IncRef()

    # Move the instance dict
    types.NULL.__dict__ = new_null.__dict__

    # Move the pointer object (excluding ob_refcnt)
    ptr_size = sizeof(c_void_p)
    memmove(
        id(types.NULL) + ptr_size,
        id(new_null) + ptr_size,
        types.NULL.__sizeof__() - ptr_size,
    )



"""Methods for managing unsafe contexts for views."""
from __future__ import annotations

from collections.abc import Generator
from contextlib import contextmanager
from functools import wraps
from typing import TYPE_CHECKING, TypeVar

from typing_extensions import ContextManager, ParamSpec, Self

from einspect.errors import UnsafeError

if TYPE_CHECKING:
    from einspect.views.view_base import View

_T = TypeVar("_T")
_P = ParamSpec("_P")


class UnsafeContext:
    _global_unsafe = False

    def __init__(self):
        super().__init__()
        self._local_unsafe = False

    @property
    def _unsafe(self) -> bool:
        """Returns True if unsafe operations are allowed."""
        return self._global_unsafe or self._local_unsafe

    def unsafe(self) -> ContextManager[Self]:
        """
        Context manager to enter an unsafe context.

        Examples:
            >>> from einspect import view
            >>> v = view(100)
            >>> with v.unsafe():
            >>>     v.size += 1

            >>> with view(100).unsafe() as v:
            >>>     v.size -= 1
        """

        def ctx() -> Generator[Self, None, None]:
            self._local_unsafe = True
            try:
                yield self
            finally:
                self._local_unsafe = False

        return contextmanager(ctx)()


@contextmanager
def global_unsafe() -> Generator[None, None, None]:
    """Context manager for global unsafe contexts."""
    UnsafeContext._global_unsafe = True
    yield None
    UnsafeContext._global_unsafe = False


def unsafe(func: _T) -> _T:
    """Decorator for unsafe methods on Views."""

    @wraps(func)
    def unsafe_call(self: View, *args: _P.args, **kwargs: _P.kwargs) -> _T:
        # noinspection PyProtectedMember
        if not self._unsafe:
            raise UnsafeError(f"Call to {func.__qualname__} requires unsafe context")
        return func(self, *args, **kwargs)

    return unsafe_call



from __future__ import annotations

import ctypes
from typing import TypeVar

from einspect.structs.py_float import PyFloatObject
from einspect.views.unsafe import unsafe
from einspect.views.view_base import View

__all__ = ("FloatView",)

_T = TypeVar("_T")


class FloatView(View[float, None, None]):
    _pyobject: PyFloatObject

    @property
    def mem_size(self):
        return object.__sizeof__(self.value)

    @property
    def value(self) -> float:
        return self._pyobject.ob_fval  # type: ignore

    @value.setter
    @unsafe
    def value(self, obj: float | ctypes.c_double) -> None:
        self._pyobject.ob_fval = obj



from __future__ import annotations

from collections.abc import Callable, MutableSequence, Sequence
from ctypes import Array, addressof, c_void_p, memmove, sizeof
from typing import Any, SupportsIndex, TypeVar, overload

from einspect.api import PTR_SIZE
from einspect.errors import UnsafeError
from einspect.structs import PyListObject, PyObject, PyTupleObject
from einspect.types import SupportsLessThan, ptr
from einspect.views.unsafe import unsafe
from einspect.views.view_base import VarView

__all__ = ("TupleView",)

_T = TypeVar("_T")


def can_resize(view: TupleView, target: int) -> bool:
    """Check if the tuple can be resized to `target` length."""
    if target <= view.size:
        return True

    current_size = view.mem_size
    delta = (target - view.size) * sizeof(c_void_p)
    return current_size + delta <= view.mem_allocated


class TupleView(VarView[tuple, None, _T], MutableSequence):
    _pyobject: PyTupleObject[_T]

    def __init__(self, obj: tuple[_T, ...] | tuple, ref: bool = True) -> None:
        """View a tuple object."""
        super().__init__(obj, ref)

    def __len__(self) -> int:
        return self.size

    @overload
    def __getitem__(self, index: SupportsIndex) -> _T:
        ...

    @overload
    def __getitem__(self, index: slice) -> tuple[_T]:
        ...

    def __getitem__(self, index: SupportsIndex | slice) -> _T | tuple[_T]:
        if isinstance(index, slice):
            start = index.start if index.start is not None else 0
            stop = index.stop if index.stop is not None else self.size
            ref = self._pyobject.GetSlice(start, stop)
            return ref.contents.into_object()
        else:
            index = index.__index__()
            ob_size = self._pyobject.ob_size
            pos_index = index if index >= 0 else ob_size + index
            py_obj = self._pyobject.GetItem(pos_index).contents
            return py_obj.into_object()

    def __setitem__(self, index: SupportsIndex | slice, value: Any) -> None:
        if isinstance(index, slice):
            # Use a temp list for the slice calculation
            temp = list(self)
            temp[index] = value
            # Check the size diff
            diff = len(temp) - self.size

            # Try to resize
            if not can_resize(self, self.size + diff) and not self._unsafe:
                raise UnsafeError(
                    "setting slice required tuple to be resized beyond current memory allocation."
                    " Enter an unsafe context to allow this."
                )
            self._pyobject.ob_size = len(temp)
            # Set the items
            for i, item in enumerate(temp):
                self[i] = item
        else:
            index = index.__index__()
            obj = PyObject.from_object(value)
            obj.IncRef()
            self.item[index] = obj.as_ref()

    @overload
    def __delitem__(self, index: SupportsIndex) -> None:
        ...

    @overload
    def __delitem__(self, index: slice) -> None:
        ...

    def __delitem__(self, index: SupportsIndex | slice) -> None:
        if isinstance(index, slice):
            r = range(self.size)[index]
            for i in reversed(r):
                del self[i]
        else:
            index = index.__index__()
            ob_size = self._pyobject.ob_size
            # Check index
            pos_index = index if index >= 0 else ob_size + index
            if pos_index >= ob_size:
                raise IndexError("tuple assignment index out of range")
            # Get the item to DecRef it
            item = self._pyobject.ob_item[index]
            if item:  # Ignore if already null pointer
                item.contents.DecRef()
            # Set the item to null
            self._pyobject.ob_item[index] = ptr[PyObject]()
            # Shift items unless this is the last item
            if pos_index != ob_size - 1:
                src = addressof(self._pyobject.ob_item[pos_index + 1])
                dst = addressof(item)
                size = (ob_size - pos_index - 1) * sizeof(c_void_p)
                memmove(dst, src, size)
            # Reduce size
            self._pyobject.ob_size -= 1

    def pop(self, index: SupportsIndex | None = None) -> _T:
        """
        Remove and return item at index (default last).

        Raises IndexError if tuple is empty or index is out of range.
        """
        index = index.__index__() if index is not None else -1
        if self.size == 0:
            raise IndexError("pop from empty tuple")
        try:
            return super().pop(index)
        except IndexError:
            raise IndexError("pop index out of range") from None

    def insert(self, index: SupportsIndex, value: _T) -> None:
        """Insert object before index."""
        index = index.__index__()
        # Normalize index
        ob_size = self._pyobject.ob_size
        pos_index = index if index >= 0 else ob_size + index
        # If below 0, use 0, if above size, use size
        norm_index = max(0, min(pos_index, ob_size))

        # Resize
        if not can_resize(self, self.size + 1) and not self._unsafe:
            raise UnsafeError(
                "insert required tuple to be resized beyond current memory allocation."
                " Enter an unsafe context to allow this."
            )
        self._pyobject.ob_size += 1

        # Shift items unless this is the last item
        if norm_index != ob_size:
            src = addressof(self._pyobject.ob_item[norm_index])
            dst = src + sizeof(c_void_p)
            size = (ob_size - norm_index) * sizeof(c_void_p)
            memmove(dst, src, size)

        # Set the item
        self._pyobject.ob_item[norm_index] = PyObject.from_object(value).as_ref()

    def sort(
        self,
        *,
        key: Callable[[str], SupportsLessThan] | None = None,
        reverse: bool = False,
    ) -> None:
        """
        Sort the tuple in ascending order and return None.

        The sort is in-place (i.e. the tuple itself is modified) and stable
        (i.e. the order of two equal elements is maintained).

        If a key function is given, apply it once to each tuple item and sort them,
        ascending or descending, according to their function values.

        The reverse flag can be set to sort in descending order.
        """
        # Skip empty
        if self._pyobject.ob_size == 0:
            return None
        temp = sorted(self._pyobject.into_object(), key=key, reverse=reverse)
        temp_obj = PyListObject.from_object(temp)
        assert temp_obj.ob_size == self._pyobject.ob_size
        # Move the references
        memmove(
            addressof(self._pyobject.ob_item),
            addressof(temp_obj.ob_item.contents),
            self._pyobject.ob_size * PTR_SIZE,
        )

    @property
    def item(self) -> Array[ptr[PyObject[_T]]]:
        """The `ob_item` array of PyObject pointers."""
        return self._pyobject.ob_item

    @item.setter
    @unsafe
    def item(
        self, value: Array[ptr[PyObject]] | Sequence[ptr[PyObject]] | Sequence[object]
    ) -> None:
        """Set the `ob_item` array of PyObject pointers."""
        if not isinstance(value, Array):
            value = [PyObject.try_from(v).with_ref().as_ref() for v in value]
        self._pyobject.ob_item = value



from __future__ import annotations

from ctypes import Structure, c_void_p, cast
from enum import IntEnum

from typing_extensions import Annotated

from einspect.api import uintptr_t
from einspect.structs.deco import struct
from einspect.structs.traits import AsRef
from einspect.types import ptr


class PyGC(IntEnum):
    """Bit flags for _gc_prev"""

    # Bit 0 is set when tp_finalize is called
    PREV_MASK_FINALIZED = 1
    # Bit 1 is set when the object is in generation which is GCed currently.
    PREV_MASK_COLLECTING = 2
    # The (N-2) most significant bits contain the real address.
    PREV_SHIFT = 2
    PREV_MASK = uintptr_t(-1).value << PREV_SHIFT


@struct
class PyGC_Head(Structure, AsRef):
    """
    Defines the PyGC_Head Structure.

    GC information is stored BEFORE the object structure.
    https://github.com/python/cpython/blob/3.11/Include/internal/pycore_gc.h#L11-L20
    """

    # Pointer to next object in the list.
    # 0 means the object is not tracked
    _gc_next: Annotated[int, uintptr_t]
    # Pointer to previous object in the list.
    # Lowest two bits are used for flags.
    _gc_prev: Annotated[int, uintptr_t]

    def Next(self) -> ptr[PyGC_Head]:
        return cast(self._gc_next, ptr[PyGC_Head])

    def Set_Next(self, item: ptr[PyGC_Head]) -> None:
        addr = cast(item, c_void_p).value
        self._gc_next = addr

    def Prev(self) -> ptr[PyGC_Head]:
        return cast(self._gc_prev & PyGC.PREV_MASK, ptr[PyGC_Head])

    def Set_Prev(self, item: ptr[PyGC_Head]) -> None:
        item = cast(item, c_void_p).value
        if item & ~PyGC.PREV_MASK != 0:
            raise ValueError("item is not valid")
        self._gc_prev = (self._gc_prev & ~PyGC.PREV_MASK) | item

    def Finalized(self) -> bool:
        return (self._gc_prev & PyGC.PREV_MASK_FINALIZED) != 0

    def Set_Finalized(self) -> None:
        self._gc_prev |= PyGC.PREV_MASK_FINALIZED



"""Parsing type hints."""
from __future__ import annotations

import ctypes
import logging
import re
import typing
from ctypes import POINTER
from typing import Any, Protocol, Sequence, TypeVar, get_origin, runtime_checkable

from typing_extensions import Self

log = logging.getLogger(__name__)

Py_ssize_t = ctypes.c_ssize_t


@runtime_checkable
class FuncPtr(Protocol):
    """Type of ctypes.pythonapi functions (ctypes._FuncPtr)."""

    restype: type | None
    argtypes: Sequence[type]

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        ...  # pragma: no cover


aliases = {
    # ctypes will cast c_ssize_t to (int)
    int: Py_ssize_t,
    # ctypes will cast py_object to (object)
    object: ctypes.py_object,
    type: ctypes.py_object,
    # convert hints of None (inspected as NoneType) back to None
    type(None): None,
}

# ctypes generics to replace
RE_PY_OBJECT = re.compile(r"^(py_object)(\[(.*)])$")
RE_POINTER = re.compile(r"^(pointer)(\[(.*)])$")
RE_ANNOTATED = re.compile(r"^(Annotated)(\[(.*)])$")


def fix_ctypes_generics(
    type_hints: dict[str, str | type], cls_name: str | None = None
) -> None:
    """
    Replace some unresolvable ctypes hints in __annotations__.

    Args:
        type_hints: The __annotations__ dict to fix.
        cls_name: The name of the class to fix.
    """
    for name, hint in type_hints.items():
        if not isinstance(hint, str):
            continue
        # For Annotated c_void_p, directly set it here to avoid name errors
        if m := RE_ANNOTATED.match(hint):
            inner = m.group(3)
            # Split from right to first ,
            _, last = inner.rsplit(",", 1)
            log.debug(last)
            if last.strip() == "c_void_p":
                log.debug("RE_ANNOTATED Set: %r -> %r", name, ctypes.c_void_p)
                type_hints[name] = ctypes.c_void_p

        # Keep py_object and discard subscript
        if m := RE_PY_OBJECT.match(hint):
            log.debug("Source: %r Match: %r", hint, m)
            base = hint.replace(m.group(2), "")
            type_hints[name] = base
            log.debug("Replacing %r with %r", hint, base)
        # For pointer, replace with POINTER
        if m := RE_POINTER.match(hint):
            # Get inner of []
            base = m.group(3)
            # Discard any other generics
            base = base.split("[")[0]
            type_hints[name] = base
            log.debug("Replacing %r with %r", hint, base)


def convert_type_hints(source: type, owner_cls: type) -> type | None:
    """Convert type hints to types usable for FuncPtr."""
    # Unpack optionals
    if hasattr(source, "__origin__"):
        if get_origin(source) == typing.Union:
            source = typing.get_args(source)[0]
            return convert_type_hints(source, owner_cls)

    # Convert TypeVar and Type to py_object
    if (type(source) is TypeVar) or get_origin(source) is type:
        return ctypes.py_object

    if source == Self:
        source = owner_cls

    # Get base of generic alias
    # noinspection PyUnresolvedReferences, PyProtectedMember
    if isinstance(source, typing._GenericAlias):
        source = get_origin(source)

    if source in aliases:
        return aliases[source]

    # Replace with a pointer type if it's a structure
    try:
        if issubclass(source, ctypes.Structure):
            return POINTER(source)
    except TypeError as e:
        raise TypeError(
            f"Failed to convert type hint ({type(source)}) {source!r} for {owner_cls!r} -> {e}"
        ) from e

    # For non-ctypes (and not None) types, substitute with py_object
    if not is_ctypes_type(source):
        source = ctypes.py_object

    return source


def is_ctypes_type(obj: Any) -> bool:
    """Return True if the object is a ctypes type."""
    try:
        ctypes.POINTER(obj)
        return True
    except TypeError:
        return False



from __future__ import annotations

from ctypes import Array, c_uint32
from typing import Iterable, TypeVar

from typing_extensions import Annotated

from einspect.structs import PyLongObject
from einspect.views.view_base import VarView

__all__ = ("IntView",)

_T = TypeVar("_T")


class IntView(VarView[int, None, None]):
    _pyobject: PyLongObject

    @property
    def digits(self) -> Array[Annotated[int, c_uint32]]:
        return self._pyobject.ob_digit

    @digits.setter
    def digits(self, value: Iterable) -> None:
        self._pyobject.ob_digit[:] = value

    @property
    def value(self) -> int:
        return self._pyobject.value

    @value.setter
    def value(self, obj: int) -> None:
        if not isinstance(obj, int):
            raise TypeError(f"Expected int, got {type(obj).__name__!r}")
        # Get a struct of the new value
        cls = type(self._pyobject)
        new_val = cls.from_object(obj)

        # The new value's ob_size must be equal or less than the current
        new_size = abs(new_val.ob_size)
        cur_size = max(abs(self._pyobject.ob_size), 1)
        if new_size > cur_size:
            raise ValueError(f"New value {obj!r} too large")

        # Copy the new value's digits into the current value
        self.digits[:new_size] = new_val.ob_digit

        # Set the new value's ob_size
        self._pyobject.ob_size = new_size



from __future__ import annotations

from contextlib import suppress
from ctypes import addressof, c_void_p, memmove
from typing import TYPE_CHECKING

from einspect.api import PTR_SIZE
from einspect.errors import UnsafeError
from einspect.structs import PyASCIIObject, PyObject, TpFlags

if TYPE_CHECKING:
    from einspect.views.view_base import View


def check_move(dst: View, src: View) -> None:
    """
    Check if a memory from `dst` to `src` is safe.

    Raises:
        UnsafeError: If the move is unsafe.
    """
    # If non-gc type into gc type, always unsafe
    if not src.is_gc() and dst.is_gc():
        raise UnsafeError(
            f"Move of non-gc type {src.type.__name__!r} into gc type"
            f" {dst.type.__name__!r} requires an unsafe context."
        )

    dst_allocated = dst.mem_allocated
    dst_offset = 0

    # Check if dst has an instance dict
    if (dst_dict := dst._pyobject.instance_dict()) is not None:
        # Set -1 to if managed dict
        if dst._pyobject.ob_type.contents.tp_flags & TpFlags.MANAGED_DICT:
            dst_offset = -1
        else:
            # If offset is positive, add this to dst_allocated
            dst_offset = addressof(dst_dict) - dst._pyobject.address
            if dst_offset > 0:
                dst_allocated = max((dst_offset + PTR_SIZE), dst_allocated)

    # Check if we need to move an instance dict
    if (src_dict := src._pyobject.instance_dict()) is not None:
        # If moving from managed -> managed, always safe
        if not (
            src._pyobject.ob_type.contents.tp_flags & TpFlags.MANAGED_DICT
            and dst_offset == -1
        ):
            src_offset = addressof(src_dict) - src._pyobject.address
            # If negative, it must match dst_offset
            neg_unsafe = src_offset < 0 and src_offset != dst_offset
            # Otherwise, needs to be within mem_allocated
            pos_unsafe = src_offset > 0 and (src_offset + PTR_SIZE) > dst_allocated
            if neg_unsafe or pos_unsafe:
                raise UnsafeError(
                    f"memory move of instance dict at offset {src_offset} from {src.type.__name__!r} "
                    f"to {dst.type.__name__!r} is out of bounds. Enter an unsafe context to allow this."
                )

    # Check if src mem_size can fit into dst mem_allocated
    if src.mem_size > dst_allocated:
        raise UnsafeError(
            f"memory move of {src.mem_size} bytes into allocated space of {dst.mem_allocated} bytes"
            " is out of bounds. Enter an unsafe context to allow this."
        )


def move(
    dst: PyObject, src: PyObject, offset: int = PTR_SIZE, inst_dict: bool = True
) -> None:
    """Move data from PyObjects `src` to `dst`."""
    # If either is a string, un-intern them
    for py_obj in (src, dst):
        if isinstance(py_obj, PyASCIIObject):
            py_obj: PyASCIIObject
            py_obj.interned = 0  # not interned
            py_obj.hash = -1  # invalidate cached hash

    # Materialize instance dicts in case we need to copy
    src_dict_ptr = None
    if inst_dict:
        with suppress(AttributeError):
            src.GetAttr("__dict__")
        with suppress(AttributeError):
            dst.GetAttr("__dict__")

        if (src_dict_ptr := src.instance_dict()) is not None:
            dict_addr = addressof(src_dict_ptr)
            # Normally we copy by offset, unless managed dict
            if not src.ob_type.contents.tp_flags & TpFlags.MANAGED_DICT:
                dict_offset = dict_addr - src.address
                memmove(
                    dst.address + dict_offset,
                    c_void_p(dict_addr),
                    PTR_SIZE,
                )

    # Move main object
    memmove(
        dst.address + offset,
        src.address + offset,
        src.mem_size - offset,
    )

    # For managed dicts, we materialize the dict after move to copy it
    if (
        inst_dict
        and src_dict_ptr is not None
        and src.ob_type.contents.tp_flags & TpFlags.MANAGED_DICT
    ):
        dst.SetAttr("__dict__", src_dict_ptr.contents.into_object())



from __future__ import annotations

from ctypes import PYFUNCTYPE, Structure, c_int, c_void_p, py_object

from einspect.api import Py_ssize_t
from einspect.structs import PyObject, struct
from einspect.types import ptr

__all__ = ("Py_buffer", "getbufferproc", "releasebufferproc")


# noinspection PyPep8Naming
@struct
class Py_buffer(Structure):
    buf: c_void_p
    obj: ptr[PyObject]  # owned reference
    len: Py_ssize_t
    itemsize: Py_ssize_t  # Py_ssize_t so it can be pointed to by strides in simple case
    readonly: c_int
    ndim: c_int
    format: c_void_p
    shape: ptr[Py_ssize_t]
    strides: ptr[Py_ssize_t]
    suboffsets: ptr[Py_ssize_t]
    internal: c_void_p


getbufferproc = PYFUNCTYPE(c_int, py_object, Py_buffer, c_int)
releasebufferproc = PYFUNCTYPE(None, py_object, Py_buffer)



from __future__ import annotations

import ctypes
import typing

# noinspection PyUnresolvedReferences, PyProtectedMember
from ctypes import _Pointer
from typing import TYPE_CHECKING, List, TypeVar, get_origin, overload

from typing_extensions import Self

from einspect.compat import Version

if TYPE_CHECKING:
    from einspect.structs import PyObject

__all__ = ("ptr", "Pointer", "Array", "NULL", "_SelfPtr", "SupportsLessThan")

_T = TypeVar("_T")

ptr = ctypes.pointer
"""
Dynamic typing alias for ctypes.pointer.

Resolves to the `_Ptr` class at runtime to allow for generic subscripting.
"""

# noinspection PyUnresolvedReferences, PyProtectedMember
Pointer = ctypes._Pointer
"""Alias to `ctypes._Pointer`."""

_SelfPtr = object()
"""Singleton object returned on ptr[Self]."""

_Null_Type = type(
    "NullType",
    (ctypes.POINTER(ctypes.c_void_p),),
    {
        "_type_": ctypes.c_void_p,
    },
)

NULL: ptr[PyObject] = _Null_Type()
"""Singleton NULL pointer of type ptr[PyObject]."""


class _Ptr(_Pointer):
    """
    Runtime alias for `ctypes.pointer` to allow generic subscripting.
    """

    def __new__(cls, *args, **kwargs):
        """Alias to `ctypes.pointer(*args, **kwargs)`"""
        return ctypes.pointer(*args, **kwargs)

    def __class_getitem__(cls, item):
        """Return a `ctypes.POINTER` of the given type."""
        # For ptr[Self], return a special object
        if item is Self:
            return _SelfPtr

        # Get base of generic alias
        # noinspection PyUnresolvedReferences, PyProtectedMember
        if isinstance(item, typing._GenericAlias):
            item = get_origin(item)

        try:
            return ctypes.POINTER(item)
        except TypeError as e:
            raise TypeError(f"{e} (During POINTER({item}))") from e


if Version.PY_3_9.above():  # pragma: no cover
    # This cannot be defined in 3.8 as ctypes.Array doesn't support subscripting
    class Array(ctypes.Array[_T]):
        """
        A typing alias for ctypes.Array for non-simple types.

        Resolves to ctypes.Array directly at runtime.
        """

        _length_ = 0
        _type_ = ctypes.c_void_p

        @overload
        def __getitem__(self, item: int) -> _T:
            ...

        @overload
        def __getitem__(self, item: slice) -> List[_T]:
            ...

        def __getitem__(self, item):
            raise NotImplementedError


# subclasses of ctypes types to avoid Structure auto casting
class void_p(ctypes.c_void_p):
    pass


class wchar_p(ctypes.c_wchar_p):
    pass


class char_p(ctypes.c_char_p):
    pass


class SupportsLessThan(typing.Protocol):  # pragma: no cover
    def __lt__(self, other: Self) -> bool:
        ...


if not TYPE_CHECKING:
    # Runtime overrides
    globals().update(
        {
            # Overwrite ctypes.pointer -> _Ptr
            "ptr": _Ptr,
            # Define Array as ctypes.Array
            "Array": ctypes.Array,
        }
    )



from __future__ import annotations

from collections.abc import Callable
from types import BuiltinFunctionType
from typing import Any, TypeVar

from einspect.structs import PyCFunctionObject
from einspect.structs.include.methodobject_h import PyMethodDef
from einspect.structs.include.object_h import vectorcallfunc
from einspect.structs.py_object import py_get, py_set
from einspect.structs.traits import IsGC
from einspect.views.unsafe import unsafe
from einspect.views.view_base import View

__all__ = ("CFunctionView",)

_T = TypeVar("_T", BuiltinFunctionType, Callable)


class CFunctionView(View[_T, None, None], IsGC):
    _pyobject: PyCFunctionObject[_T]

    def __init__(self, obj: _T, ref: bool = False) -> None:
        super().__init__(obj, ref)

    @property
    def ml(self) -> PyMethodDef:
        return self._pyobject.m_ml.contents

    # noinspection PyPropertyDefinition, PyMethodParameters
    @property
    def self(__self):
        return py_get(__self._pyobject.m_self)

    # noinspection PyPropertyDefinition, PyMethodParameters
    @self.setter
    @unsafe
    def self(__self, value):
        """Passed as 'self' arg to the C func."""
        py_set(__self._pyobject.m_self, value)

    @property
    def module(self) -> str:
        """The __module__ attribute, can be anything."""
        return self._pyobject.m_module.contents.into_object()

    @module.setter
    def module(self, value: str) -> None:
        py_set(self._pyobject.m_module, value)

    @property
    def weakreflist(self) -> Any:
        """Weak reference list"""
        return self._pyobject.m_weakreflist.contents.into_object()

    @weakreflist.setter
    def weakreflist(self, value: Any) -> None:
        py_set(self._pyobject.m_weakreflist, value)

    @property
    def vectorcall(self) -> vectorcallfunc:
        """Vectorcall function"""
        return self._pyobject.vectorcall

    @vectorcall.setter
    @unsafe
    def vectorcall(self, value: Callable) -> None:
        self._pyobject.vectorcall = vectorcallfunc(value)



"""Python version compatibility constructs."""
from __future__ import annotations

import sys
import typing
from collections import abc
from dataclasses import dataclass
from enum import Enum
from typing import Generic, NoReturn, TypeVar

__all__ = ("Version", "RequiresPythonVersion", "python_req", "abc")

if sys.version_info > (3, 8):
    abc = typing  # noqa: F401, F811


class Version(Enum):
    """Python version constants."""

    PY_3_7 = (3, 7)
    PY_3_8 = (3, 8)
    PY_3_9 = (3, 9)
    PY_3_10 = (3, 10)
    PY_3_11 = (3, 11)
    PY_3_12 = (3, 12)

    def above(self, or_eq: bool = True) -> bool:
        """Return whether the current version is above this version."""
        if or_eq:
            return sys.version_info >= self.value
        return sys.version_info > self.value

    def below(self, or_eq: bool = False) -> bool:
        """Return whether the current version is below this version."""
        if or_eq:
            return sys.version_info <= self.value
        return sys.version_info < self.value

    def req_above(self, or_eq: bool = True) -> RequiresPythonVersion | None:
        """Return None if above a python version, else a RequiresPythonVersion instance."""
        return None if self.above(or_eq) else RequiresPythonVersion(self)

    def req_below(self, or_eq: bool = False) -> RequiresPythonVersion | None:
        """Return None if below a python version, else a RequiresPythonVersion instance."""
        return None if self.below(or_eq) else RequiresPythonVersion(self)


_V = TypeVar("_V", bound=Version)


@dataclass
class RequiresPythonVersion(Generic[_V]):
    version: _V

    def __msg__(self) -> str:
        return f"Requires Python {self.version.value[0]}.{self.version.value[1]}"

    def __call__(self, *args, **kwargs) -> NoReturn:
        raise RuntimeError(self.__msg__())

    def __getitem__(self, _item) -> RequiresPythonVersion[_V]:
        return self


def python_req(version: _V) -> RequiresPythonVersion[_V] | None:
    """
    Returns None if equal or greater than current Python version.

    If lower, returns an 'RequiresPythonVersion' instance.
    """
    if sys.version_info < version.value:  # type: ignore
        return RequiresPythonVersion(version)
    return None



from __future__ import annotations

from collections.abc import Callable, MutableSequence
from ctypes import Array
from typing import Iterable, SupportsIndex, TypeVar, Union, overload

from einspect.api import Py_ssize_t
from einspect.errors import UnsafeError
from einspect.structs import (
    Kind,
    PyASCIIObject,
    PyCompactUnicodeObject,
    PyUnicodeObject,
    State,
)
from einspect.types import SupportsLessThan
from einspect.views.unsafe import unsafe
from einspect.views.view_base import REF_DEFAULT, View

__all__ = ("StrView",)

_T = TypeVar("_T")


def _check_resize(v: StrView, target: int, method: str) -> None:
    """
    Check if the string can be resized to `target` size.

    Raises:
        UnsafeError: If the string cannot be resized, and not in an unsafe context.
    """
    # See if resizing is safe
    if target > v.mem_allocated and not v._unsafe:
        raise UnsafeError(
            f"string {method} required {target} bytes, beyond current"
            f" allocated memory of {v.mem_allocated} bytes."
            " Enter an unsafe context to allow this."
        )


def _str_move(v: StrView, target: str) -> None:
    """Move the string to a new location."""
    # Since the struct type might change, use a memmove
    target_v = StrView(target, ref=False)
    with target_v.unsafe():
        target_v.move_to(v)
    v._narrow_type()
    # On modifications, unintern the string
    v._pyobject.interned = 0


class StrView(View[str, None, None], MutableSequence):
    _pyobject: Union[PyASCIIObject, PyCompactUnicodeObject, PyUnicodeObject]

    def __init__(self, obj: str, ref: bool = REF_DEFAULT) -> None:
        """View a string object."""
        super().__init__(obj, ref)
        # Start with PyASCIIObject, check if we can use a more specific type
        self._narrow_type()

    def _narrow_type(self) -> None:
        """Narrow to a more specific unicode type if possible."""
        if self._pyobject.compact:
            if self._pyobject.ascii:
                self._pyobject = self._pyobject.astype(PyASCIIObject)
            else:
                self._pyobject = self._pyobject.astype(PyCompactUnicodeObject)
        else:
            self._pyobject = self._pyobject.astype(PyUnicodeObject)

    def __len__(self) -> int:
        return self._pyobject.GetLength()

    @overload
    def __getitem__(self, index: int) -> str:
        ...

    @overload
    def __getitem__(self, index: slice) -> str:
        ...

    def __getitem__(self, index: int | slice) -> str:
        return str.__getitem__(self.base, index)

    @overload
    def __setitem__(self, index: int, value: str) -> None:
        ...

    @overload
    def __setitem__(self, index: slice, value: Iterable[str]) -> None:
        ...

    def __setitem__(self, index: int | slice, value: str | Iterable[str]) -> None:
        if not isinstance(index, slice):
            index = index.__index__()
            if len(value) > 1:
                raise ValueError(
                    "cannot set integer indices with string of length > 1, use a slice instead."
                )
        # Use a temp list for the slice calculation
        temp = list(self._pyobject.into_object())
        temp[index] = value
        # Combine to str
        target = "".join(temp)

        # See if resizing is safe, then move the string
        _check_resize(self, target.__sizeof__(), "setitem")
        _str_move(self, target)

    @overload
    def __delitem__(self, index: int) -> None:
        ...

    @overload
    def __delitem__(self, index: slice) -> None:
        ...

    def __delitem__(self, index: int) -> None:
        # Use a temp list for the slice calculation
        temp = list(self._pyobject.into_object())
        del temp[index]
        # Combine to str
        target = "".join(temp)
        # See if resizing is safe, then move the string
        _check_resize(self, target.__sizeof__(), "delitem")
        _str_move(self, target)

    def insert(self, index: SupportsIndex, value: str) -> None:
        """Insert object before index."""
        index = index.__index__()
        # Use a temp list for the slice calculation
        temp = list(self._pyobject.into_object())
        temp.insert(index, value)
        # Combine to str
        target = "".join(temp)
        # See if resizing is safe, then move the string
        _check_resize(self, target.__sizeof__(), "insert")
        _str_move(self, target)

    def remove(self, value: str) -> None:
        """
        Remove first occurrence of substring value.

        Raises ValueError if the value is not present.
        """
        # Equivalent to str.replace(value, "", 1)
        target = str.replace(self._pyobject.into_object(), value, "", 1)
        # See if resizing is safe (should always be safe but check anyway)
        _check_resize(self, target.__sizeof__(), "remove")
        _str_move(self, target)

    def sort(
        self,
        *,
        key: Callable[[str], SupportsLessThan] | None = None,
        reverse: bool = False,
    ) -> None:
        """
        Sort the string in ascending order and return None.

        The sort is in-place (i.e. the string itself is modified) and stable
        (i.e. the order of two equal elements is maintained).

        If a key function is given, apply it once to each character and sort them,
        ascending or descending, according to their function values.

        The reverse flag can be set to sort in descending order.
        """
        # Skip empty
        if not self._pyobject.length:
            return None
        temp = list(self._pyobject.into_object())
        temp.sort(key=key, reverse=reverse)
        # Combine to str
        target = "".join(temp)
        # See if resizing is safe, then move the string
        _check_resize(self, target.__sizeof__(), "sort")
        _str_move(self, target)

    @property
    def length(self) -> int:
        return self._pyobject.length

    @length.setter
    def length(self, value: int) -> None:
        self._pyobject.length = value

    @property
    def hash(self) -> int:
        return self._pyobject.hash

    @hash.setter
    @unsafe
    def hash(self, value: int) -> None:
        self._pyobject.hash = value

    @property
    def interned(self) -> State:
        return State(self._pyobject.interned)

    @interned.setter
    def interned(self, value: State | int) -> None:
        self._pyobject.interned = value

    @property
    def kind(self) -> Kind:
        return Kind(self._pyobject.kind)

    @kind.setter
    def kind(self, value: Kind | int) -> None:
        self._pyobject.kind = value

    @property
    def buffer(self) -> Array[Py_ssize_t]:
        return self._pyobject.buffer



"""Traits and mixins for structs."""
from ctypes import Structure, addressof, pointer

from typing_extensions import Self

from einspect.types import ptr


class AsRef:
    """Mixin for an as_ref method."""

    def as_ref(self) -> ptr[Self]:
        """Return a pointer to the Structure."""
        return pointer(self)  # type: ignore


class Display:
    """Mixin for displaying Structures."""

    def __repr__(self: Structure) -> str:
        """
        Return a string representation of the Structure.
        The address shown is the address of the Structure, not the object.
        """
        return f"<{self.__class__.__name__} at {addressof(self):#04x}>"


class IsGC:
    """Mixin for Structures that have a GC_Head."""

    _is_gc_ = True



from __future__ import annotations

from ctypes import (
    POINTER,
    Structure,
    _SimpleCData,
    addressof,
    c_char,
    c_int8,
    c_int16,
    c_int32,
    c_int64,
    c_uint8,
    c_uint32,
    c_uint64,
    pythonapi,
)
from typing import Dict, TypeVar

from typing_extensions import Annotated

from einspect.protocols.delayed_bind import bind_api
from einspect.structs.deco import struct
from einspect.structs.py_object import Fields, PyObject
from einspect.structs.traits import Display, IsGC
from einspect.types import Array, ptr

__all__ = ("PyDictObject",)


_KT = TypeVar("_KT")
_VT = TypeVar("_VT")


@struct
class PyDictKeysObject(Structure, Display):
    """
    Defines a DictKeysObject Structure.

    https://github.com/python/cpython/blob/3.11/Include/internal/pycore_dict.h#L87-L125
    """

    dk_refcnt: int
    dk_log2_size: Annotated[int, c_uint8]
    dk_log2_index_bytes: Annotated[int, c_uint8]
    dk_kind: Annotated[int, c_uint8]
    dk_version: Annotated[int, c_uint32]
    dk_usable: int
    dk_nentries: int
    # Actual hash table of dk_size entries. It holds indices in dk_entries,
    # or DKIX_EMPTY(-1) or DKIX_DUMMY(-2).
    # Indices must be: 0 <= indice < USABLE_FRACTION(dk_size).
    #
    # The size in bytes of an indice depends on dk_size:
    # - 1 byte if dk_size <= 0xff (char*)
    # - 2 bytes if dk_size <= 0xffff (int16_t*)
    # - 4 bytes if dk_size <= 0xffffffff (int32_t*)
    # - 8 bytes otherwise (int64_t*)
    #
    # Dynamically sized, SIZEOF_VOID_P is minimum.
    _dk_indices: c_int8 * 0

    @property
    def dk_indices(self) -> Array[int]:
        items_addr = addressof(self._dk_indices)
        item_type = self._dk_indices_type()[0]
        arr = item_type * self._dk_size
        return arr.from_address(items_addr)

    @property
    def _dk_size(self) -> int:
        return 1 << self.dk_log2_size

    def _dk_indices_type(self) -> tuple[_SimpleCData, str]:
        if self._dk_size <= 0xFF:
            return c_char, "c_char"
        elif self._dk_size <= 0xFFFF:
            return c_int16, "c_int16"
        elif self._dk_size <= 0xFFFFFFFF:  # pragma: no branch
            return c_int32, "c_int32"
        return c_int64, "c_int64"  # pragma: no cover

    def _format_fields_(self) -> Fields:
        indice_type, indice_name = self._dk_indices_type()
        return {
            "dk_refcnt": "Py_ssize_t",
            "dk_log2_size": "c_uint8",
            "dk_log2_index_bytes": "c_uint8",
            "dk_kind": "c_uint8",
            "dk_version": "c_uint32",
            "dk_usable": "Py_ssize_t",
            "dk_nentries": "Py_ssize_t",
            "dk_indices": f"Array[{indice_name}]",
        }


@struct
class PyDictValues(Structure, Display):
    values: ptr[PyObject]


@struct
class PyDictObject(PyObject[dict, _KT, _VT], IsGC):
    """
    Defines a PyDictObject Structure.

    https://github.com/python/cpython/blob/3.11/Include/cpython/dictobject.h
    """

    # Number of items in the dictionary
    ma_used: int
    # Dictionary version: globally unique, changes on modification
    ma_version_tag: Annotated[int, c_uint64]
    ma_keys: ptr[PyDictKeysObject]
    # If ma_values is NULL, the table is "combined": keys and values
    # are stored in ma_keys. Otherwise, keys are stored in ma_keys
    # and values are stored in ma_values
    ma_values: ptr[PyDictValues]

    def _format_fields_(self) -> Fields:
        return {
            **super()._format_fields_(),
            "ma_used": "Py_ssize_t",
            "ma_version_tag": "c_uint64",
            "ma_keys": "*PyDictKeysObject",
            "ma_values": "*PyDictValues",
        }

    @bind_api(pythonapi["PyDict_GetItem"])
    def GetItem(self, key: _KT) -> POINTER(PyObject):
        """Return a pointer to the value object at key, or NULL if the key is not found."""

    @bind_api(pythonapi["PyDict_SetItem"])
    def SetItem(self, key: _KT, val: _VT) -> int:
        """
        Set a value to a given key.

        This function does not steal a reference to val.

        Returns:
            0 on success or -1 on failure.

        Raises:
            TypeError: if the key is not hashable.
        """

    @bind_api(pythonapi["PyDict_DelItem"])
    def DelItem(self, key: _KT) -> int:
        """
        Remove the entry with key.

        Returns:
            0 on success or -1 on failure.

        Raises:
            TypeError: if the key is not hashable.
            KeyError: if the key is not present.
        """

    @classmethod
    def from_object(cls, obj: Dict[_KT, _VT]) -> PyDictObject[_KT, _VT]:
        """Create a PyDictObject from an object."""
        return super().from_object(obj)  # type: ignore



from __future__ import annotations

from ctypes import Array
from typing import TypeVar

from einspect.api import Py_ssize_t
from einspect.compat import abc
from einspect.structs.py_dict import PyDictKeysObject, PyDictObject
from einspect.types import ptr
from einspect.views.unsafe import unsafe
from einspect.views.view_base import REF_DEFAULT, View

__all__ = ("DictView",)

_KT = TypeVar("_KT")
_VT = TypeVar("_VT")


class DictView(View[dict, _KT, _VT], abc.MutableMapping[_KT, _VT]):
    _pyobject: PyDictObject[_KT, _VT]

    def __init__(self, obj: dict[_KT, _VT], ref: bool = REF_DEFAULT) -> None:
        super().__init__(obj, ref)

    def __len__(self) -> int:
        return self.used

    def __iter__(self) -> abc.Iterator[_KT]:
        return self.base.__iter__()

    def __getitem__(self, key: _KT) -> _VT:
        """Get an item from the dictionary. Equivalent to dict[key]."""
        obj = self._pyobject.GetItem(key)
        if not obj:  # Null pointer
            raise KeyError(key)
        return obj.contents.into_object()

    def __setitem__(self, key: _KT, value: _VT) -> None:
        if self._pyobject.SetItem(key, value) < 0:
            raise SystemError("Call to PyDict_SetItem failed.")  # pragma: no cover
        return None

    def __delitem__(self, key: _KT) -> None:
        if self._pyobject.DelItem(key) < 0:
            raise SystemError("Call to PyDict_DelItem failed.")  # pragma: no cover
        return None

    @property
    def used(self) -> int:
        return self._pyobject.ma_used

    @used.setter
    @unsafe
    def used(self, value: int) -> None:
        self._pyobject.ma_used = value

    @property
    def version_tag(self) -> int:
        return self._pyobject.ma_version_tag

    @version_tag.setter
    @unsafe
    def version_tag(self, value: int) -> None:
        self._pyobject.ma_version_tag = value

    @property
    def ma_keys(self) -> ptr[PyDictKeysObject]:
        return self._pyobject.ma_keys

    @ma_keys.setter
    @unsafe
    def ma_keys(self, value: Array[Py_ssize_t]) -> None:
        self._pyobject.ma_keys = value

    @property
    def ma_values(self) -> Array[Py_ssize_t]:
        return self._pyobject.ma_values  # type: ignore

    @ma_values.setter
    @unsafe
    def ma_values(self, value: Array[Py_ssize_t]) -> None:
        self._pyobject.ma_values = value



"""View formatted info implementation."""
from __future__ import annotations

import ctypes
from ctypes import Array, Structure, cast
from string import Template
from typing import TYPE_CHECKING, Any

from einspect.structs.py_object import PyObject

if TYPE_CHECKING:
    from einspect.views.view_base import View

BASE = Template("$struct_name (at $address):")
INDENTS_COUNT = 3
INDENTS = " " * INDENTS_COUNT

# Display transforms for types
DISP_TRANSFORMS = {
    type: lambda c: c.__name__,
}


def indent(s: str) -> str:
    """Indent a string."""
    # For single line strings, just add indents
    if "\n" not in s:
        return INDENTS + s
    # For multi-line, add indents to each line
    return "\n".join(INDENTS + line for line in s.splitlines())


class Formatter:
    def __init__(self, types: bool = True, arr_max: int | None = None):
        """
        Initialize a Formatter.

        Args:
            types: Whether to include type hints.
            arr_max: The max number of elements to show for Array types.
        """
        self.types = types
        self.arr_max = arr_max

    def format_value(self, obj: Any) -> str:
        """Format a value."""
        # Array: format as list
        if isinstance(obj, Array):
            # If there is a max, limit the number of elements
            diff = 0
            if self.arr_max is not None and len(obj) > self.arr_max:
                diff = len(obj) - self.arr_max
                obj = obj[: self.arr_max]
            res = list(map(self.format_value, obj))

            # For structures (first char is \n), add \n to last and indent
            if res and res[0].startswith("{ ") and res[0].endswith(" }"):
                sep = f",\n{INDENTS}"
                text = f"{sep.join(res)}{', ...' if diff > 0 else ''}"
                return f"[\n{INDENTS}{text}\n]"

            return f"[{', '.join(res)}{', ...' if diff > 0 else ''}]"
        # For pointers, get the value
        # noinspection PyUnresolvedReferences, PyProtectedMember
        if isinstance(obj, ctypes._Pointer):
            val = self.format_value(obj.contents) if obj else "NULL"
            if val.startswith("[") and val.endswith("]"):
                return f"&{val}"
            return f"&[{val}]"
        # For PyObject, get the object
        if isinstance(obj, PyObject):
            obj.IncRef()
            return self.format_value(obj.into_object())
        # For ctypes Structures, do multi-line formatting
        if isinstance(obj, Structure):
            return self.format_structure(obj, short=True)
        # Other cases
        try:
            return DISP_TRANSFORMS[type(obj)](obj)
        except KeyError:
            # Fallback to repr
            return repr(obj)

    def format_attr(
        self,
        struct: PyObject,
        attr: str,
        hint: str | tuple[str, type],
        type_hints: bool | None = None,
    ) -> str:
        value = getattr(struct, attr)
        type_cast = None
        if isinstance(hint, tuple):
            hint, type_cast = hint

        use_hints = self.types if type_hints is None else type_hints
        type_str = f": {hint}" if use_hints else ""

        # If cast_to provided
        if type_cast is not None:
            value = cast(value, type_cast)
            value = getattr(value, "value", value)

        res = f"{attr}{type_str} = {self.format_value(value)}"
        return res

    # noinspection PyProtectedMember
    def format_structure(self, struct: Structure, short: bool | None = None) -> str:
        # Short mode looks like
        # { ob_refcnt = 1, ob_type = &[object] }
        lines = []
        # Auto short if below 5
        if short is None:
            short = len(struct._format_fields_()) < 5
        # Get attributes
        for attr, type_hint in struct._format_fields_().items():
            text = self.format_attr(
                struct=struct,
                attr=attr,
                hint=type_hint,
                # If short, don't show type hints
                type_hints=(None if not short else False),
            )
            if not short:
                text = indent(text)
            lines.append(text)
        if short:
            return f"{{ {', '.join(lines)} }}"
        return "\n" + "\n".join(lines) + "\n"

    # noinspection PyProtectedMember
    def format_view(self, v: View) -> str:
        """Format a display string for a View."""
        lines = []
        # Get struct
        struct = v._pyobject
        struct_name = struct.__class__.__name__
        address = f"0x{struct.address:x}"
        # Make line 1
        lines.append(BASE.substitute(struct_name=struct_name, address=address))
        # Get attributes
        for attr, type_hint in struct._format_fields_().items():
            lines.append(indent(self.format_attr(struct, attr, type_hint)))

        return "\n".join(lines)



from __future__ import annotations

from ctypes import (
    POINTER,
    Structure,
    c_char,
    c_char_p,
    c_uint,
    c_ulong,
    c_void_p,
    pointer,
    py_object,
    pythonapi,
)
from typing import Any, Type, TypeVar

from typing_extensions import Annotated, Self

from einspect.protocols import bind_api
from einspect.structs.deco import struct
from einspect.structs.include.descrobject_h import PyGetSetDef, PyMemberDef
from einspect.structs.include.methodobject_h import PyMethodDef
from einspect.structs.include.object_h import *
from einspect.structs.py_object import PyObject, PyVarObject
from einspect.structs.slots_map import get_slot
from einspect.types import ptr

__all__ = ("PyTypeObject",)

_T = TypeVar("_T")

DEFAULT = object()


# noinspection PyPep8Naming
@struct
class PyTypeObject(PyVarObject[_T, None, None]):
    """
    Defines a PyTypeObject Structure.

    https://github.com/python/cpython/blob/3.11/Doc/includes/typestruct.h
    """

    tp_name: Annotated[bytes, c_char_p]
    # For allocation
    tp_basicsize: int
    tp_itemsize: int
    # Methods to implement standard operations
    tp_dealloc: destructor
    tp_vectorcall_offset: int
    tp_getattr: getattrfunc
    tp_setattr: setattrfunc
    # formerly known as tp_compare (Python 2) or tp_reserved (Python 3)
    tp_as_async: ptr[PyAsyncMethods]

    tp_repr: reprfunc

    # Method suites for standard classes
    tp_as_number: ptr[PyNumberMethods]
    tp_as_sequence: ptr[PySequenceMethods]
    tp_as_mapping: ptr[PyMappingMethods]

    # More standard operations (here for binary compatibility)
    tp_hash: hashfunc
    tp_call: ternaryfunc
    tp_str: reprfunc
    tp_getattro: getattrofunc
    tp_setattro: setattrofunc

    # Functions to access object as input/output buffer
    tp_as_buffer: ptr[PyBufferProcs]

    # Flags to define presence of optional/expanded features
    tp_flags: Annotated[int, c_ulong]

    tp_doc: Annotated[bytes, c_char_p]  # Documentation string

    # Assigned meaning in release 2.0
    # call function for all accessible objects
    tp_traverse: traverseproc

    tp_clear: inquiry  # delete references to contained objects

    # Assigned meaning in release 2.1
    # rich comparisons
    tp_richcompare: richcmpfunc

    tp_weaklistoffset: int  # weak reference enabler

    # Iterators
    tp_iter: getiterfunc
    tp_iternext: iternextfunc

    # Attribute descriptor and subclassing stuff
    tp_methods: ptr[PyMethodDef]
    tp_members: ptr[PyMemberDef]
    tp_getset: ptr[PyGetSetDef]

    # Strong reference on a heap type, borrowed reference on a static type
    tp_base: pointer[Self]
    tp_dict: ptr[PyObject]
    tp_descr_get: descrgetfunc
    tp_descr_set: descrsetfunc
    tp_dictoffset: int
    tp_init: initproc
    tp_alloc: allocfunc
    tp_new: newfunc
    tp_free: freefunc  # Low-level free-memory routine
    tp_is_gc: inquiry  # For PyObject_IS_GC
    tp_bases: ptr[PyObject]
    tp_mro: ptr[PyObject]  # method resolution order
    tp_cache: ptr[PyObject]
    tp_subclasses: c_void_p  # for static builtin types this is an index
    tp_weaklist: ptr[PyObject]
    tp_del: destructor

    # Type attribute cache version tag. Added in version 2.6
    tp_version_tag: Annotated[int, c_uint]

    tp_finalize: destructor
    tp_vectorcall: vectorcallfunc

    # bitset of which type-watchers care about this type
    tp_watched: c_char

    def __new__(cls, __obj: Type[_T] = DEFAULT, **kwargs):
        """
        Create a new PyTypeObject.

        One positional argument can be passed to create a PyTypeObject from a type object.
        Otherwise, keyword arguments of fields can be provided.

        Args:
            __obj: A Python object to create a PyTypeObject from.
            **kwargs: Fields to create a PyObject from.
        """
        if __obj is not DEFAULT:
            # If already a PyObject
            if isinstance(__obj, PyObject):
                return cls.from_address(__obj.address)
            # Use from_object
            return cls.from_object(__obj)

        # Base case
        return Structure.__new__(cls)

    def __repr__(self) -> str:
        """Return a string representation of the PyTypeObject."""
        cls_name = f"{self.__class__.__name__}"
        type_name = self.tp_name.decode()
        return f"<{cls_name}[{type_name}] at {self.address:#04x}>"

    @classmethod
    def from_object(cls, obj: Type[_T]) -> PyTypeObject[Type[_T]]:
        return super().from_object(obj)  # type: ignore

    def setattr_safe(self, name: str, value: Any) -> None:
        """Set an attribute on the type object. Uses custom overrides if available."""
        # Resolve the slot into an attr name, if any
        if (slot := get_slot(name)) is not None:
            # Override c_char_p types
            if type_object_fields.get(slot.name) is c_char_p:
                setattr(self, slot.name, value.encode())
                return
            # Override ptr[PyObject] types
            if type_object_fields.get(slot.name) == ptr[PyObject]:
                setattr(self, slot.name, PyObject.from_object(value).as_ref())
                return
        self.SetAttr(name, value)

    def is_gc(self) -> bool:
        """
        Return True if the type has GC support.

        https://docs.python.org/3/c-api/type.html#c.PyType_IS_GC
        https://github.com/python/cpython/blob/3.11/Include/objimpl.h#L160-L161
        """
        return bool(self.tp_flags & TpFlags.HAVE_GC)

    @bind_api(pythonapi["PyType_Modified"])
    def Modified(self) -> None:
        """Mark the type as modified."""

    @bind_api(pythonapi["_PyObject_New"])
    def New(self) -> ptr[PyObject]:
        """Create a new object of the type."""

    @bind_api(pythonapi["_PyObject_GC_New"])
    def GC_New(self) -> ptr[PyObject]:
        """Create a new object of the type with GC support."""

    @bind_api(pythonapi["_PyObject_NewVar"])
    def NewVar(self, nitems: int) -> ptr[PyVarObject]:
        """Create a new variable object of the type."""

    @bind_api(pythonapi["_PyObject_GC_NewVar"])
    def GC_NewVar(self, nitems: int) -> ptr[PyVarObject]:
        """Create a new variable object of the type with GC support."""


# Mapping of CField name to type
# noinspection PyProtectedMember
type_object_fields = {name: t for name, t, *_ in PyTypeObject._fields_}


# Patch PyTypeObject back into PyObject
def _patch_py_object():
    # noinspection PyProtectedMember
    fields = PyObject._fields_
    fields[1] = ("ob_type", POINTER(PyTypeObject))
    offset = object.__basicsize__ + tuple.__itemsize__ * 3
    py_object.from_address(id(PyObject.ob_type) + offset).value = POINTER(PyTypeObject)


_patch_py_object()



"""Decorator protocols for binding class properties."""
from __future__ import annotations

import ctypes
import inspect
import logging
from collections.abc import Callable
from functools import partial
from inspect import signature
from types import MethodType
from typing import TypeVar, get_type_hints

from einspect.protocols.type_parse import (
    FuncPtr,
    convert_type_hints,
    fix_ctypes_generics,
)

log = logging.getLogger(__name__)

_F = TypeVar("_F")
_R = TypeVar("_R")
_CT = TypeVar("_CT", bound=ctypes.Structure)


def bind_api(py_api: FuncPtr) -> Callable[[_F], _F]:
    """
    Decorator to bind a ctypes FuncPtr function to a class.

    Type hints of the decorated function are used to determine the
    argtypes and restype of the FuncPtr.
    """
    return partial(delayed_bind, py_api)


# noinspection PyPep8Naming
class delayed_bind(property):
    def __init__(self, py_api: FuncPtr, func: _F):
        super().__init__()
        # Use __func__ if staticmethod
        if isinstance(func, staticmethod):
            func = func.__func__
        self.func = func
        self.__doc__ = func.__doc__
        self.py_api = py_api
        self.attrname: str | None = None
        self.restype = None
        self.argtypes = None
        self.func_set = False

    def __repr__(self):
        return f"<{self.__class__.__name__} property {self.attrname!r}>"

    def __set_name__(self, owner: type, name: str) -> None:
        if self.attrname is None:
            self.attrname = name
        elif name != self.attrname:  # pragma: no cover
            raise TypeError(
                "Cannot assign the same bind to two different names "
                f"({self.attrname!r} and {name!r})."
            )

    def _get_defining_type_hints(self, owner_cls: type) -> tuple[list[type], type]:
        """Return the type hints for the attribute we're bound to, or None if it's not defined."""
        fix_ctypes_generics(self.func.__annotations__)
        # Get the function type hints
        hints = get_type_hints(self.func)
        log.debug(
            "[%s.%s()] Type hints: %s",
            owner_cls.__qualname__,
            self.attrname,
            hints,
        )
        res_t = hints.pop("return", None)
        arg_t = list(hints.values())

        # Disallow any missing type hints
        if None in arg_t or res_t is None:
            raise TypeError(
                "Cannot resolve bind function type hints. "
                "Please provide them explicitly."
            )

        res_t = convert_type_hints(res_t, owner_cls)

        # Insert current class type as first argument
        # If there is a "self" parameter
        if signature(self.func).parameters.get("self") and "self" not in hints:
            # Here we want to insert the actual class the function is defined
            # Not subclasses (in case a subclass gets the first call)
            def_cls = _get_defining_class_of_bound_method(self.func, owner_cls)
            log.debug("Found defining class: %s of %s", def_cls, self.func)
            arg_t.insert(0, def_cls)

        arg_t = [convert_type_hints(t, owner_cls) for t in arg_t]

        return arg_t, res_t

    # noinspection PyMethodOverriding
    def __get__(self, instance: object | None, owner_cls: type[_CT]) -> _F:
        if self.attrname is None:
            raise TypeError(
                "Cannot use bind instance without calling __set_name__ on it."
            )

        if not self.func_set:
            argtypes, restype = self._get_defining_type_hints(owner_cls)
            self.py_api.argtypes = argtypes
            self.py_api.restype = restype
            self.func_set = True
            if log.isEnabledFor(logging.DEBUG):
                log.debug(
                    "[%s.%s()] Set func: (%s) -> %r",
                    owner_cls.__name__,
                    self.attrname,
                    ", ".join(
                        repr(x.__name__) if x is not None else "None" for x in argtypes
                    ),
                    self.py_api.restype.__name__
                    if self.py_api.restype is not None
                    else "None",
                )

        # Called as class method, return directly without binding
        if instance is None:
            return self.py_api  # type: ignore

        try:
            cache = instance.__dict__
        except AttributeError:
            raise TypeError(
                "delayed_bind requires class to support __dict__."
            ) from None

        bound_func = MethodType(self.py_api, instance)

        try:
            cache[self.attrname] = bound_func
        except TypeError:
            msg = (
                f"The '__dict__' attribute on {type(instance).__name__!r} instance "
                f"does not support item assignment for caching {self.attrname!r} property."
            )
            raise TypeError(msg) from None

        return bound_func


def _get_defining_class_of_bound_method(method, current_cls) -> type:
    """Get defining class of a bound method."""
    for cls in inspect.getmro(current_cls):
        if method.__name__ in cls.__dict__:
            return cls
    raise ValueError(f"Failed to find defining class of {method!r}")



from __future__ import annotations

from types import MappingProxyType
from typing import TypeVar

from einspect.compat import abc
from einspect.structs import PyDictObject
from einspect.structs.mapping_proxy import MappingProxyObject
from einspect.views.unsafe import unsafe
from einspect.views.view_base import REF_DEFAULT, View

__all__ = ("MappingProxyView",)

_KT = TypeVar("_KT")
_VT = TypeVar("_VT")


class MappingProxyView(View[MappingProxyType, _KT, _VT], abc.MutableMapping[_KT, _VT]):
    _pyobject: MappingProxyObject[_KT, _VT]

    def __init__(
        self, obj: MappingProxyType[_KT, _VT], ref: bool = REF_DEFAULT
    ) -> None:
        super().__init__(obj, ref)

    def __len__(self) -> int:
        return dict.__len__(self.mapping)

    def __iter__(self) -> abc.Iterator[_KT]:
        return dict.__iter__(self.mapping)

    def __getitem__(self, __k: _KT) -> _VT:
        return dict.__getitem__(self.mapping, __k)

    def __setitem__(self, __k: _KT, __v: _VT) -> None:
        return dict.__setitem__(self.mapping, __k, __v)

    def __delitem__(self, __v: _KT) -> None:
        return dict.__delitem__(self.mapping, __v)

    @property
    def mapping(self) -> dict[_KT, _VT]:
        """Return `MappingProxyObject.mapping`, casted to `dict`."""
        return self._pyobject.mapping.contents.into_object()

    @mapping.setter
    @unsafe
    def mapping(self, value: dict[_KT, _VT]) -> None:
        self._pyobject.mapping = PyDictObject.from_object(value).with_ref().as_ref()



"""Proxy for retrieving original methods and slot wrappers of types."""
from __future__ import annotations

from typing import Any, Type, TypeVar

_T = TypeVar("_T")
MISSING = object()

obj_getattr = object.__getattribute__
type_hash = type.__hash__
str_eq = str.__eq__
dict_setdefault = dict.setdefault
dict_contains = dict.__contains__
dict_get = dict.get
dict_getitem = dict.__getitem__

_slots_cache: dict[type, dict[str, Any]] = {}


def add_cache(type_: type, name: str, method: Any):
    """Add a method to the cache."""
    type_methods = dict_setdefault(_slots_cache, type_, {})
    # Only allow adding once, ignore if already added
    dict_setdefault(type_methods, name, method)


def in_cache(type_: type, name: str) -> bool:
    """Return True if the method is in the cache."""
    type_methods = dict_setdefault(_slots_cache, type_, {})
    return dict_contains(type_methods, name)


def get_cache(type_: type, name: str) -> Any:
    """Get the method from the type in cache."""
    type_methods = dict_setdefault(_slots_cache, type_, {})
    return dict_getitem(type_methods, name)


class orig:
    """
    Proxy to access a type's original attributes.

    Attributes on this object will mirror original attributes
    on the type before modification by @impl or TypeView.
    """

    def __new__(cls, type_: Type[_T]) -> Type[_T]:
        obj = object.__new__(cls)
        obj.__type = type_
        return obj  # type: ignore

    def __repr__(self) -> str:
        return f"orig({self.__type.__name__})"

    def __call__(self, *args, **kwargs):
        """Call the original type."""
        return self.__type(*args, **kwargs)

    def __getattribute__(self, name: str):
        """Get an attribute from the original type."""
        # Overrides
        _type = obj_getattr(self, "_orig__type")
        if str_eq(name, "_orig__type"):
            return _type

        # Check if the attribute is cached
        try:
            return get_cache(_type, name)
        except KeyError:
            pass
        # Get the attribute from the original type and cache it
        attr = getattr(_type, name)
        add_cache(_type, name, attr)
        return attr



"""Structures for CPython objects."""
from __future__ import annotations

import ctypes
import warnings
from contextlib import suppress
from ctypes import POINTER, Structure, c_void_p, pythonapi
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generic,
    List,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

from typing_extensions import Annotated, Self

from einspect.api import PTR_SIZE, address, align_size
from einspect.compat import Version, python_req
from einspect.protocols.delayed_bind import bind_api
from einspect.protocols.type_parse import is_ctypes_type
from einspect.structs.deco import struct
from einspect.structs.py_gc import PyGC_Head
from einspect.structs.traits import AsRef, IsGC
from einspect.types import ptr

if TYPE_CHECKING:
    from einspect.structs import PyTypeObject

__all__ = ("PyObject", "PyVarObject", "Fields", "py_get", "py_set")

Fields = Dict[str, Union[str, Tuple[str, Type]]]

_T = TypeVar("_T")
_KT = TypeVar("_KT")
_VT = TypeVar("_VT")
_ST = TypeVar("_ST", bound=Structure)

DEFAULT = object()


def py_get(obj_ptr: ptr[PyObject]) -> object | None:
    """
    Get a PyObject pointer value.

    If the pointer is NULL, return None.
    """
    return obj_ptr.contents.into_object() if obj_ptr else None


def py_set(obj_ptr: ptr[PyObject], value: object | PyObject | ptr[PyObject]) -> None:
    """
    Set a PyObject pointer value on a View.

    If the attribute exists and points to a PyObject, it will be DecRef'd.
    The value will be IncRef'd before being set.
    """
    # Get and DecRef current
    if obj_ptr:
        obj_ptr.contents.DecRef()
    # Set new
    obj_ptr.contents = PyObject.try_from(value).with_ref()


@struct
class PyObject(Structure, AsRef, Generic[_T, _KT, _VT]):
    """Defines a base PyObject Structure."""

    ob_refcnt: int
    ob_type: Annotated[ptr[PyTypeObject[Type[_T]]], c_void_p]

    _fields_: List[Union[Tuple[str, type], Tuple[str, type, int]]]

    @overload
    def __new__(cls, __obj: _T):
        ...

    @overload
    def __new__(cls, *, ob_refcnt: int = 1, ob_type: ptr[PyTypeObject], **kwargs):
        ...

    def __new__(cls, __obj: _T = DEFAULT, **kwargs):
        """
        Create a new PyObject.

        One positional argument can be passed to create a PyObject from an object.
        Otherwise, keyword arguments of fields can be provided.

        Args:
            __obj: A Python object to create a PyObject from.
            **kwargs: Fields to create a PyObject from.
        """
        from einspect.structs.py_type import PyTypeObject

        # Base case
        if __obj is DEFAULT and not kwargs:
            return super().__new__(cls)
        if __obj is not DEFAULT:
            # If already a PyObject
            if isinstance(__obj, PyObject):
                return cls.from_address(__obj.address)
            # Use from_object
            return cls.from_object(__obj)
        # Check kwargs
        if "ob_type" not in kwargs:
            raise TypeError("Missing required keyword-argument field 'ob_type'")

        new_ob_size: int = kwargs.get("ob_size", 0).__index__()
        new_ob_type = PyTypeObject.try_from(kwargs["ob_type"])

        if issubclass(cls, IsGC):
            res = (
                new_ob_type.GC_NewVar(new_ob_size)
                if issubclass(cls, PyVarObject)
                else new_ob_type.GC_New()
            )
        else:
            res = (
                new_ob_type.NewVar(new_ob_size)
                if issubclass(cls, PyVarObject)
                else new_ob_type.New()
            )
        return res.contents.astype(cls)

    @property
    def mem_size(self) -> int:
        """Return the size of the PyObject in memory."""
        return ctypes.sizeof(self)

    @property
    def address(self) -> int:
        """Return the address of the PyObject."""
        return ctypes.addressof(self)

    def __eq__(self, other: Self) -> bool:
        """Return True if the PyObject is equal in address to the other."""
        if not isinstance(other, PyObject):
            return NotImplemented
        return self.address == other.address

    def __repr__(self) -> str:
        """Return a string representation of the PyObject."""
        cls_name = f"{self.__class__.__name__}"
        # For generic PyObjects, add the type name
        if self.__class__ in (PyObject, PyVarObject):
            with suppress(ValueError):
                obj_type = self.ob_type.contents
                cls_name += f"[{obj_type.tp_name.decode()}]"
        return f"<{cls_name} at {self.address:#04x}>"

    def _format_fields_(self) -> Fields:
        """
        Return an attribute mapping for info display.

        Returns:
            Dict mapping of field to type-hint or (type-hint, cast type)
        """
        return {"ob_refcnt": "Py_ssize_t", "ob_type": "*PyTypeObject"}

    @classmethod
    def from_object(cls, obj: _T) -> Self:
        """Create a PyObject from an object."""
        return cls.from_address(address(obj))

    @classmethod
    def from_gc(cls, gc: PyGC_Head) -> Self:
        """Create a PyObject from a PyGC_Head struct."""
        addr = ctypes.addressof(gc) + ctypes.sizeof(PyGC_Head)
        return cls.from_address(addr)

    @classmethod
    def try_from(cls, obj_or_ptr: PyObject | ptr[PyObject] | object) -> Self:
        """Create a PyObject from a PyObject, pointer to a PyObject, or object."""
        # noinspection PyUnresolvedReferences
        if isinstance(obj_or_ptr, ctypes._Pointer):
            obj = obj_or_ptr.contents
        else:
            obj = obj_or_ptr
        # For ctypes types
        if is_ctypes_type(type(obj)):
            if isinstance(obj, PyObject):
                return cls.from_object(obj.into_object())
            # raise if not a PyObject
            raise TypeError(f"Cannot create PyObject from {obj_or_ptr!r}")
        # For Python objects
        return cls.from_object(obj)

    def into_object(self) -> _T:
        """Cast the PyObject into a Python object."""
        py_obj = ctypes.cast(self.as_ref(), ctypes.py_object)
        return py_obj.value

    def astype(self, dtype: Type[_ST]) -> _ST:
        """Cast the PyObject into another PyObject type."""
        return dtype.from_address(self.address)

    def with_ref(self, n: int = 1) -> Self:
        """Increment the reference count of the PyObject by n. Return self."""
        self.ob_refcnt += n
        return self

    def is_gc(self) -> bool:
        """
        Returns True if the object implements the GC protocol.

        The object cannot be tracked by the garbage collector if False.

        https://github.com/python/cpython/blob/3.11/Include/internal/pycore_object.h#L209-L216
        """
        return (type_obj := self.ob_type.contents).is_gc() and (
            not type_obj.tp_is_gc or type_obj.tp_is_gc(self.into_object())
        )

    def as_gc(self) -> PyGC_Head:
        """Return the PyGC_Head struct of this object."""
        addr = self.address - ctypes.sizeof(PyGC_Head)
        return PyGC_Head.from_address(addr)  # type: ignore

    def gc_is_tracked(self) -> bool:
        """Return True if the PyObject is currently tracked by the GC."""
        return self.as_gc()._gc_next != 0

    def gc_may_be_tracked(self) -> bool:
        """
        Return True if the PyObject may be tracked by
        the GC in the future, or already is.

        https://github.com/python/cpython/blob/3.11/Include/internal/pycore_gc.h#L28-L32
        """
        if not self.is_gc():
            return False
        if self.ob_type.contents.into_object() is tuple:
            return self.gc_is_tracked()
        return True

    def instance_dict(self) -> ptr[PyObject[dict, Any, Any]] | None:
        """
        Return the instance dict of the PyObject.

        An offset override can be set by `__st_dictoffset__`. It should be
        relative to the address of the PyObject.

        https://docs.python.org/3/c-api/typeobj.html#c.PyTypeObject.tp_dictoffset
        """
        # Get the tp_dictoffset of the type
        offset = self.ob_type.contents.tp_dictoffset
        # If 0, the type does not have a dict
        if offset == 0:
            return None
        # For -1, the type uses the 3.12 Managed Dict feature
        if offset == -1:
            # Check that the flag is set
            from einspect.structs import TpFlags

            if not self.ob_type.contents.tp_flags & TpFlags.MANAGED_DICT:
                raise RuntimeError(
                    "type has a __dictoffset__ of -1, but tp_flags does not have TpFlags.MANAGED_DICT"
                )
            # Materialize the dict
            inst_dict = self.GetAttr("__dict__")
            return PyObject.from_object(inst_dict).as_ref()
        # For > 0, start from the address of the PyObject
        if offset > 0:
            addr = self.address + offset
        # For < 0, start after the struct
        else:
            # If not a PyVarObject, use __basic_size__ instead
            if getattr(self, "ob_size", None) is None:
                size = self.ob_type.contents.tp_basicsize
            else:
                size = align_size(self.mem_size, PTR_SIZE)
                # Increase size by pointer size since mem_size at this point
                # excludes the instance dict (unlike __sizeof__)
                size += PTR_SIZE
            addr = self.address + size + offset
        # Return the pointer
        return POINTER(PyObject).from_address(addr)

    @bind_api(python_req(Version.PY_3_10) or pythonapi["Py_NewRef"])
    def NewRef(self) -> object:
        """Returns new reference of the PyObject."""

    @bind_api(pythonapi["Py_IncRef"])
    def IncRef(self) -> None:
        """Increment the reference count of the PyObject."""

    @bind_api(pythonapi["Py_DecRef"])
    def DecRef(self) -> None:
        """Decrement the reference count of the PyObject."""

    @bind_api(pythonapi["PyObject_GetAttr"])
    def GetAttr(self, name: str) -> object:
        """Return the attribute of the PyObject."""

    @bind_api(pythonapi["PyObject_SetAttr"])
    def SetAttr(self, name: str, value: object) -> int:
        """Set the attribute of the PyObject."""


# We don't want this visible to type-checking but `PyObject.__init__`
# needs to ignore the `__obj` positional argument
orig_init = PyObject.__init__  # type: ignore


def _PyObject__init__(self, *args, **kwargs) -> None:
    if len(args) > 1:
        warnings.warn(
            "Positional arguments to PyObject.__init__ are not supported, use keywords instead.",
            UserWarning,
        )
    return orig_init(self, **kwargs)  # type: ignore


PyObject.__init__ = _PyObject__init__


@struct
class PyVarObject(PyObject[_T, _KT, _VT]):
    """
    Defines a base PyVarObject Structure.

    https://github.com/python/cpython/blob/3.11/Include/object.h#L109-L112
    """

    ob_size: int

    def _format_fields_(self) -> Fields:
        return {**super()._format_fields_(), "ob_size": "Py_ssize_t"}



from __future__ import annotations

from typing import ContextManager

import einspect._patch
from einspect.type_orig import orig
from einspect.types import NULL, ptr
from einspect.views.factory import view
from einspect.views.unsafe import global_unsafe
from einspect.views.view_type import impl

# Runtime patches
einspect._patch.run()

__all__ = ("view", "unsafe", "impl", "orig", "ptr", "NULL")

__version__ = "0.5.7.post1"

unsafe: ContextManager[None] = global_unsafe



from __future__ import annotations

from ctypes import Structure
from typing import Generic, TypeVar

from typing_extensions import Annotated

from einspect.api import Py_hash_t
from einspect.structs.deco import struct
from einspect.structs.py_object import Fields, PyObject
from einspect.structs.traits import AsRef, IsGC
from einspect.types import ptr

_T = TypeVar("_T")

PySet_MINSIZE = 8
"""https://github.com/python/cpython/blob/3.11/Include/cpython/setobject.h#L18"""


@struct
class SetEntry(Structure, AsRef, Generic[_T]):
    key: ptr[PyObject[_T, None, None]]
    hash: Annotated[int, Py_hash_t]  # noqa: A003

    @staticmethod
    def _format_fields_() -> Fields:
        return {
            "key": "*PyObject",
            "hash": "Py_hash_t",
        }


@struct
class PySetObject(PyObject[set, None, _T], AsRef, IsGC):
    """
    Defines a PySetObject Structure.

    https://github.com/python/cpython/blob/3.11/Include/cpython/setobject.h#L36-L59
    """

    fill: int  # Number active and dummy entries
    used: int  # Number active entries
    mask: int  # The table contains mask + 1 slots
    table: ptr[SetEntry[_T]]
    hash: Annotated[int, Py_hash_t]  # noqa: A003
    finger: int
    smalltable: SetEntry * PySet_MINSIZE
    weakreflist: ptr[PyObject]

    def _format_fields_(self) -> Fields:
        return {
            **super()._format_fields_(),
            "fill": "Py_ssize_t",
            "used": "Py_ssize_t",
            "mask": "Py_ssize_t",
            "table": "*SetEntry",
            "hash": "Py_hash_t",
            "finger": "Py_ssize_t",
            "smalltable": "Array[SetEntry]",
            "weakreflist": "*PyObject",
        }

    @classmethod
    def from_object(cls, obj: set[_T]) -> PySetObject[_T]:
        return cls.from_address(id(obj))



from __future__ import annotations

from collections.abc import Callable
from types import CodeType, FunctionType
from typing import Any, Dict

from einspect.compat import Version
from einspect.structs import PyFunctionObject
from einspect.structs.include.object_h import vectorcallfunc
from einspect.structs.py_object import py_get, py_set
from einspect.structs.traits import IsGC
from einspect.views.unsafe import unsafe
from einspect.views.view_base import View

__all__ = ("FunctionView",)


class FunctionView(View[FunctionType, None, None], IsGC):
    _pyobject: PyFunctionObject

    def __init__(self, obj: FunctionType | Callable, ref: bool = False) -> None:
        super().__init__(obj, ref)

    @property
    def globals(self) -> Dict[str, Any]:
        return self._pyobject.globals.contents.into_object()

    @globals.setter
    @unsafe
    def globals(self, value: Dict[str, Any]) -> None:
        py_set(self._pyobject.globals, value)

    @property
    def builtins(self) -> Dict[str, Any]:
        if Version.PY_3_10.below():
            raise AttributeError(
                "PyFunctionObject does not have builtins below Python 3.10"
            )
        return self._pyobject.builtins.contents.into_object()

    @builtins.setter
    @unsafe
    def builtins(self, value: Dict[str, Any]) -> None:
        if Version.PY_3_10.below():
            raise AttributeError(
                "PyFunctionObject does not have builtins below Python 3.10"
            )
        py_set(self._pyobject.builtins, value)

    @property
    def name(self) -> str:
        return self._pyobject.name.contents.into_object()

    @name.setter
    def name(self, value: str) -> None:
        py_set(self._pyobject.name, value)

    @property
    def qualname(self) -> str:
        return self._pyobject.qualname.contents.into_object()

    @qualname.setter
    def qualname(self, value: str) -> None:
        py_set(self._pyobject.qualname, value)

    @property
    def code(self) -> CodeType:
        return self._pyobject.code.contents.into_object()

    @code.setter
    def code(self, value: CodeType) -> None:
        py_set(self._pyobject.code, value)

    @property
    def defaults(self) -> tuple[Any, ...] | None:
        return py_get(self._pyobject.defaults)

    @defaults.setter
    def defaults(self, value: tuple[Any, ...] | None) -> None:
        py_set(self._pyobject.defaults, value)

    @property
    def kwdefaults(self) -> Dict[str, Any] | None:
        return py_get(self._pyobject.kwdefaults)

    @kwdefaults.setter
    def kwdefaults(self, value: Dict[str, Any] | None) -> None:
        py_set(self._pyobject.kwdefaults, value)

    @property
    def closure(self) -> tuple[Any, ...] | None:
        return py_get(self._pyobject.closure)

    @closure.setter
    @unsafe
    def closure(self, value: tuple[Any, ...] | None) -> None:
        py_set(self._pyobject.closure, value)

    @property
    def doc(self) -> str | None:
        return self._pyobject.func_doc.contents.into_object()

    @doc.setter
    def doc(self, value: str | None) -> None:
        py_set(self._pyobject.func_doc, value)

    @property
    def dict(self) -> Dict[str, Any] | None:
        return py_get(self._pyobject.func_dict)

    @dict.setter
    def dict(self, value: Dict[str, Any] | None) -> None:
        py_set(self._pyobject.func_dict, value)

    @property
    def weakreflist(self) -> list[Any]:
        return self._pyobject.func_weakreflist.contents.into_object()

    @weakreflist.setter
    @unsafe
    def weakreflist(self, value: list[Any]) -> None:
        py_set(self._pyobject.func_weakreflist, value)

    @property
    def module(self) -> str:
        return self._pyobject.func_module.contents.into_object()

    @module.setter
    def module(self, value: str) -> None:
        py_set(self._pyobject.func_module, value)

    @property
    def annotations(self) -> Dict[str, Any] | None:
        return py_get(self._pyobject.func_annotations)

    @annotations.setter
    def annotations(self, value: Dict[str, Any] | None) -> None:
        py_set(self._pyobject.func_annotations, value)

    @property
    def vectorcall(self) -> Callable:
        return self._pyobject.vectorcall.contents.into_object()

    @vectorcall.setter
    @unsafe
    def vectorcall(self, value: Callable) -> None:
        self._pyobject.vectorcall = vectorcallfunc(value)

    @property
    def version(self) -> int:
        if Version.PY_3_11.below():
            raise AttributeError(
                "PyFunctionObject does not have version below Python 3.11"
            )
        return self._pyobject.func_version

    @version.setter
    @unsafe
    def version(self, value: int) -> None:
        if Version.PY_3_11.below():
            raise AttributeError(
                "PyFunctionObject does not have version below Python 3.11"
            )
        self._pyobject.func_version = value



# Mapping of dunder names to their corresponding slot paths.
# https://docs.python.org/3/c-api/typeobj.html
from __future__ import annotations

from dataclasses import dataclass
from typing import Final, Sequence, Union

from einspect.structs.include.object_h import (
    PyAsyncMethods,
    PyMappingMethods,
    PyNumberMethods,
    PySequenceMethods,
)


@dataclass
class Slot:
    name: str
    ptr_type: type | None = None

    @property
    def parts(self) -> list[str]:
        return self.name.split(".")

    def __getitem__(self, slot: str | Slot):
        slot_name = slot.name if isinstance(slot, Slot) else slot
        return Slot(f"{self.name}.{slot_name}", self.ptr_type)


SlotsLike = Union[str, Slot, Sequence[Union[str, Slot]]]

tp_as_async = Slot("tp_as_async", PyAsyncMethods)
tp_as_number = Slot("tp_as_number", PyNumberMethods)
tp_as_sequence = Slot("tp_as_sequence", PySequenceMethods)
tp_as_mapping = Slot("tp_as_mapping", PyMappingMethods)

SLOTS_MAIN: Final[dict[str, SlotsLike]] = {
    "__name__": "tp_name",
    "__repr__": "tp_repr",
    "__hash__": "tp_hash",
    "__call__": "tp_call",
    "__str__": "tp_str",
    "__getattribute__": "tp_getattro",
    "__getattr__": "tp_getattro",
    "__setattr__": "tp_setattro",
    "__delattr__": "tp_setattro",
    "__doc__": "tp_doc",
    "__lt__": "tp_richcompare",
    "__le__": "tp_richcompare",
    "__eq__": "tp_richcompare",
    "__ne__": "tp_richcompare",
    "__gt__": "tp_richcompare",
    "__ge__": "tp_richcompare",
    "__iter__": "tp_iter",
    "__next__": "tp_iternext",
    "__base__": "tp_base",
    "__dict__": "tp_dict",
    "__get__": "tp_descr_get",
    "__set__": "tp_descr_set",
    "__delete__": "tp_descr_set",
    "__init__": "tp_init",
    "__new__": "tp_new",
    "__bases__": "tp_bases",
    "__mro__": "tp_mro",
    "__subclasses__": "tp_subclasses",
    "__del__": "tp_finalize",
}

SLOTS_SUB_ASYNC: Final[dict[str, SlotsLike]] = {
    "__await__": tp_as_async["am_await"],
    "__aiter__": tp_as_async["am_aiter"],
    "__anext__": tp_as_async["am_anext"],
}

SLOTS_NUMBERS: Final[dict[str, SlotsLike]] = {
    "__add__": tp_as_number["nb_add"],
    "__radd__": tp_as_number["nb_add"],
    "__iadd__": tp_as_number["nb_inplace_add"],
    "__sub__": tp_as_number["nb_subtract"],
    "__rsub__": tp_as_number["nb_subtract"],
    "__isub__": tp_as_number["nb_inplace_subtract"],
    "__mul__": tp_as_number["nb_multiply"],
    "__rmul__": tp_as_number["nb_multiply"],
    "__imul__": tp_as_number["nb_inplace_multiply"],
    "__mod__": tp_as_number["nb_remainder"],
    "__rmod__": tp_as_number["nb_remainder"],
    "__imod__": tp_as_number["nb_inplace_remainder"],
    "__divmod__": tp_as_number["nb_divmod"],
    "__rdivmod__": tp_as_number["nb_divmod"],
    "__pow__": tp_as_number["nb_power"],
    "__rpow__": tp_as_number["nb_power"],
    "__ipow__": tp_as_number["nb_inplace_power"],
    "__neg__": tp_as_number["nb_negative"],
    "__pos__": tp_as_number["nb_positive"],
    "__abs__": tp_as_number["nb_absolute"],
    "__bool__": tp_as_number["nb_bool"],
    "__invert__": tp_as_number["nb_invert"],
    "__lshift__": tp_as_number["nb_lshift"],
    "__rlshift__": tp_as_number["nb_lshift"],
    "__ilshift__": tp_as_number["nb_inplace_lshift"],
    "__rshift__": tp_as_number["nb_rshift"],
    "__rrshift__": tp_as_number["nb_rshift"],
    "__irshift__": tp_as_number["nb_inplace_rshift"],
    "__and__": tp_as_number["nb_and"],
    "__rand__": tp_as_number["nb_and"],
    "__iand__": tp_as_number["nb_inplace_and"],
    "__xor__": tp_as_number["nb_xor"],
    "__rxor__": tp_as_number["nb_xor"],
    "__ixor__": tp_as_number["nb_inplace_xor"],
    "__or__": tp_as_number["nb_or"],
    "__ror__": tp_as_number["nb_or"],
    "__ior__": tp_as_number["nb_inplace_or"],
    "__int__": tp_as_number["nb_int"],
    "__float__": tp_as_number["nb_float"],
    "__floordiv__": tp_as_number["nb_floor_divide"],
    "__ifloordiv__": tp_as_number["nb_inplace_floor_divide"],
    "__truediv__": tp_as_number["nb_true_divide"],
    "__itruediv__": tp_as_number["nb_inplace_true_divide"],
    "__index__": tp_as_number["nb_index"],
    "__matmul__": tp_as_number["nb_matrix_multiply"],
    "__imatmul__": tp_as_number["nb_inplace_matrix_multiply"],
}

SLOTS_SEQUENCE: Final[dict[str, SlotsLike]] = {
    "__len__": tp_as_sequence["sq_length"],
    "__add__": tp_as_sequence["sq_concat"],
    "__mul__": tp_as_sequence["sq_repeat"],
    "__getitem__": tp_as_sequence["sq_item"],
    "__setitem__": tp_as_sequence["sq_ass_item"],
    "__delitem__": tp_as_sequence["sq_ass_item"],
    "__contains__": tp_as_sequence["sq_contains"],
    "__iadd__": tp_as_sequence["sq_inplace_concat"],
    "__imul__": tp_as_sequence["sq_inplace_repeat"],
}

SLOTS_MAPPING: Final[dict[str, SlotsLike]] = {
    "__len__": tp_as_mapping["mp_length"],
    "__getitem__": tp_as_mapping["mp_subscript"],
    "__setitem__": tp_as_mapping["mp_ass_subscript"],
    "__delitem__": tp_as_mapping["mp_ass_subscript"],
}

# Set of all dunder slot names
SLOTS_NAMES: set[str] = {
    *SLOTS_MAIN.keys(),
    *SLOTS_SUB_ASYNC.keys(),
    *SLOTS_NUMBERS.keys(),
    *SLOTS_SEQUENCE.keys(),
    *SLOTS_MAPPING.keys(),
}

SLOTS = [
    SLOTS_MAIN,
    SLOTS_SUB_ASYNC,
    SLOTS_NUMBERS,
    SLOTS_SEQUENCE,
    SLOTS_MAPPING,
]


def get_slot(name: str) -> Slot | None:
    # Iterate the list to find a match
    for slots in SLOTS:
        if name in slots:
            res = slots[name]
            # Make a Slot if we got a str
            if isinstance(res, str):
                return Slot(res)
            return res
    return None



from __future__ import annotations


class MovedError(ReferenceError):
    """Unreachable weak referenced object."""


class UnsafeError(RuntimeError):
    """Unsafe operation without unsafe context."""


class DroppedReferenceError(RuntimeError):
    """Raised on access to a dropped object's reference."""


class DroppedReference:
    """Dropped reference to an object."""

    __slots__ = ("__parent_cls",)

    def __init__(self, parent_cls: type) -> None:
        """Initialize the dropped reference."""
        self.__parent_cls = parent_cls

    def __getattr__(self, item):
        raise DroppedReferenceError(
            f"Reference to {self.__parent_cls.__name__!r} has been dropped, "
            "cannot access attributes."
        ) from None



from __future__ import annotations

from ctypes import PYFUNCTYPE, Structure, c_char_p, c_int, c_void_p, py_object

from typing_extensions import Annotated

from einspect.structs import struct

__all__ = ("getter", "setter", "PyMemberDef", "PyGetSetDef")

getter = PYFUNCTYPE(py_object, py_object, py_object)
setter = PYFUNCTYPE(c_int, py_object, py_object, c_void_p)


@struct
class PyMemberDef(Structure):
    name: c_char_p
    type: Annotated[int, c_int]
    offset: int
    flags: Annotated[int, c_int]
    doc: c_char_p


@struct
class PyGetSetDef(Structure):
    name: c_char_p
    get: getter
    set: setter
    doc: c_char_p
    closure: c_void_p



from __future__ import annotations

import ctypes

from einspect.structs.deco import struct
from einspect.structs.py_object import PyObject


@struct
class PyFloatObject(PyObject):
    """
    Defines a PyFloatObject Structure.

    https://github.com/python/cpython/blob/3.11/Include/cpython/floatobject.h
    """

    ob_fval: ctypes.c_double



from __future__ import annotations

import sys

from einspect.structs.deco import struct
from einspect.structs.py_object import PyObject, PyVarObject  # Req import first
from einspect.structs.py_type import PyTypeObject, TpFlags
from einspect.structs.py_long import PyLongObject  # Req before PyBoolObject
from einspect.structs.py_bool import PyBoolObject
from einspect.structs.py_float import PyFloatObject
from einspect.structs.py_list import PyListObject
from einspect.structs.py_tuple import PyTupleObject
from einspect.structs.py_set import PySetObject, SetEntry
from einspect.structs.py_cfunction import PyCFunctionObject

# Req before PyMappingProxyObject
from einspect.structs.py_dict import PyDictObject, PyDictKeysObject, PyDictValues
from einspect.structs.mapping_proxy import MappingProxyObject

# Compatibility override for py_function
if sys.version_info > (3, 11):
    import einspect.structs.py_function as py_function
    from einspect.structs.py_function import PyFunctionObject
elif sys.version_info > (3, 10):
    import einspect._compat.py_function_3_10 as py_function
    from einspect._compat.py_function_3_10 import PyFunctionObject
else:
    import einspect._compat.py_function_3_9 as py_function
    from einspect._compat.py_function_3_9 import PyFunctionObject

# Compatibility override for Python 3.12
if sys.version_info < (3, 12):
    import einspect.structs.py_unicode as py_unicode
    from einspect.structs.py_unicode import (
        PyUnicodeObject,
        PyCompactUnicodeObject,
        PyASCIIObject,
        Kind,
        State,
    )
else:
    import einspect._compat.py_unicode_3_12 as py_unicode
    from einspect._compat.py_unicode_3_12 import (
        PyUnicodeObject,
        PyCompactUnicodeObject,
        PyASCIIObject,
        Kind,
        State,
    )

__all__ = (
    "struct",
    "PyObject",
    "PyVarObject",
    "PyTypeObject",
    "TpFlags",
    "PyTupleObject",
    "PyListObject",
    "PyLongObject",
    "PyBoolObject",
    "PyFloatObject",
    "PyUnicodeObject",
    "PyFunctionObject",
    "PyCFunctionObject",
    "State",
    "Kind",
    "PyCompactUnicodeObject",
    "py_function",
    "py_unicode",
    "PyASCIIObject",
    "PyDictObject",
    "PyDictKeysObject",
    "PyDictValues",
    "MappingProxyObject",
    "PySetObject",
    "SetEntry",
)



"""Function factory to create views for objects."""
from __future__ import annotations

from collections.abc import Callable
from types import BuiltinFunctionType, FunctionType, MappingProxyType
from typing import Any, Final, TypeVar, overload

from einspect.views import CFunctionView
from einspect.views.view_base import REF_DEFAULT, View
from einspect.views.view_bool import BoolView
from einspect.views.view_dict import DictView
from einspect.views.view_float import FloatView
from einspect.views.view_function import FunctionView
from einspect.views.view_int import IntView
from einspect.views.view_list import ListView
from einspect.views.view_mapping_proxy import MappingProxyView
from einspect.views.view_set import SetView
from einspect.views.view_str import StrView
from einspect.views.view_tuple import TupleView
from einspect.views.view_type import TypeView

__all__ = ("view",)

VIEW_TYPES: Final[dict[type, type[View]]] = {
    object: View,
    type: TypeView,
    int: IntView,
    bool: BoolView,
    float: FloatView,
    str: StrView,
    list: ListView,
    tuple: TupleView,
    set: SetView,
    dict: DictView,
    MappingProxyType: MappingProxyView,
    FunctionType: FunctionView,
    BuiltinFunctionType: CFunctionView,
}
"""Mapping of (type): (view class)."""

# Collection generics
_KT = TypeVar("_KT")
_VT = TypeVar("_VT")

# TypeView
_Type = TypeVar("_Type", bound=type)

# Base case
_T = TypeVar("_T")


@overload
def view(obj: int, ref: bool = REF_DEFAULT) -> IntView:
    ...


@overload
def view(obj: bool, ref: bool = REF_DEFAULT) -> BoolView:
    ...


@overload
def view(obj: list[_VT], ref: bool = REF_DEFAULT) -> ListView[_VT]:
    ...


@overload
def view(
    obj: MappingProxyType[_KT, _VT], ref: bool = REF_DEFAULT
) -> MappingProxyView[_KT, _VT]:
    ...


@overload
def view(obj: dict[_KT, _VT], ref: bool = REF_DEFAULT) -> DictView[_KT, _VT]:
    ...


@overload
def view(obj: set[_VT], ref: bool = REF_DEFAULT) -> SetView[_VT]:
    ...


@overload
def view(obj: tuple[_VT, ...], ref: bool = REF_DEFAULT) -> TupleView[_VT]:
    ...


@overload
def view(obj: tuple, ref: bool = REF_DEFAULT) -> TupleView[Any]:
    ...


@overload
def view(obj: str, ref: bool = REF_DEFAULT) -> StrView:
    ...


@overload
def view(obj: float, ref: bool = REF_DEFAULT) -> FloatView:
    ...


@overload
def view(obj: _Type, ref: bool = REF_DEFAULT) -> TypeView[_Type]:
    ...


# Ideally we'd have separate overloads for builtin vs user-defined functions,
# but typeshed only defines builtin functions as Callable.
@overload
def view(
    obj: BuiltinFunctionType | FunctionType | Callable, ref: bool = REF_DEFAULT
) -> FunctionView | CFunctionView:
    ...


def view(obj, ref: bool = REF_DEFAULT):
    """
    Create a view onto a Python object.

    Args:
        obj: The object to view.
        ref: If True, hold a reference to the object.

    Returns:
        A view onto the object.
    """
    obj_type = type(obj)
    if obj_type in VIEW_TYPES:
        return VIEW_TYPES[obj_type](obj, ref=ref)

    # Fallback to subclasses
    for obj_type, view_type in reversed([*VIEW_TYPES.items()]):
        if isinstance(obj, obj_type):
            return view_type(obj, ref=ref)

    # Shouldn't reach here since we will at least match isinstance object
    raise TypeError(f"Cannot create view for {obj_type}")  # pragma: no cover



from __future__ import annotations

import ctypes
from collections.abc import Sequence
from ctypes import addressof, c_uint32, sizeof

from typing_extensions import Annotated

from einspect.api import seq_to_array
from einspect.structs.deco import struct
from einspect.structs.py_object import Fields, PyVarObject
from einspect.types import Array


@struct
class PyLongObject(PyVarObject[int, None, None]):
    """
    Defines a PyLongObject Structure.

    https://github.com/python/cpython/blob/3.11/Include/cpython/longintrepr.h#L79-L82
    """

    _ob_digit_0: ctypes.c_uint32 * 0

    def _format_fields_(self) -> Fields:
        return {**super()._format_fields_(), "ob_digit": "Array[c_uint32]"}

    @property
    def mem_size(self) -> int:
        """Return the size of the PyObject in memory."""
        # Need to add size(uint32) * ob_size to our base size
        base = super().mem_size
        # use size 1 if ob_size is 0 due to allocation
        size = max(1, abs(self.ob_size))
        return base + ctypes.sizeof(c_uint32) * size

    @property
    def ob_digit(self) -> Array[Annotated[int, c_uint32]]:
        # Note PyLongObject uses the sign bit of ob_size to indicate its own sign
        # ob_size < 0 means the number is negative
        # ob_size > 0 means the number is positive
        # ob_size == 0 means the number is zero
        # The true size of the ob_digit array is abs(ob_size)
        items_addr = ctypes.addressof(self._ob_digit_0)
        size = max(abs(int(self.ob_size)), 1)
        return (c_uint32 * size).from_address(items_addr)  # type: ignore

    @ob_digit.setter
    def ob_digit(self, value: Array[int | c_uint32] | Sequence[int]) -> None:
        arr = seq_to_array(value, c_uint32)
        ctypes.memmove(
            addressof(self._ob_digit_0),
            addressof(arr),
            sizeof(arr),
        )

    @property
    def value(self) -> int:
        if self.ob_size == 0:
            return 0
        digit: int  # noqa: F842
        val = sum(digit * 1 << (30 * i) for i, digit in enumerate(self.ob_digit))
        size: int = self.ob_size  # type: ignore
        return val * (-1 if size < 0 else 1)



from __future__ import annotations

from types import MappingProxyType
from typing import TypeVar

from einspect.structs.deco import struct
from einspect.structs.py_dict import PyDictObject
from einspect.structs.py_object import Fields, PyObject
from einspect.structs.traits import IsGC
from einspect.types import ptr

__all__ = ("MappingProxyObject",)


_KT = TypeVar("_KT")
_VT_co = TypeVar("_VT_co", covariant=True)


@struct
class MappingProxyObject(PyObject[MappingProxyType, _KT, _VT_co], IsGC):
    """
    Defines a mappingproxyobject Structure.

    https://github.com/python/cpython/blob/3.11/Objects/descrobject.c#L1027-L1030
    """

    mapping: ptr[PyDictObject[_KT, _VT_co]]

    def _format_fields_(self) -> Fields:
        return {
            **super()._format_fields_(),
            "mapping": "*PyDictObject",
        }

    @classmethod
    def from_object(
        cls, obj: MappingProxyType[_KT, _VT_co]
    ) -> MappingProxyObject[MappingProxyType, _KT, _VT_co]:
        """Create a MappingProxyObject from an object."""
        return super().from_object(obj)  # type: ignore



from __future__ import annotations

from einspect.api import address
from einspect.structs import PyBoolObject
from einspect.views.view_int import IntView

__all__ = ("BoolView",)


class BoolView(IntView):
    _pyobject: PyBoolObject

    @property
    def mem_size(self) -> int:
        # If used on the singletons, use their
        addr = self._pyobject.address
        if addr == address(True):
            return True.__sizeof__()
        elif addr == address(False):
            return False.__sizeof__()
        return super().mem_size



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['einspect',
 'einspect._compat',
 'einspect.protocols',
 'einspect.structs',
 'einspect.structs.include',
 'einspect.views']

package_data = \
{'': ['*']}

install_requires = \
['typing-extensions>=4.4.0,<5.0.0']

setup_kwargs = {
    'name': 'einspect',
    'version': '0.5.7.post1',
    'description': 'Extended Inspect - view and modify memory structs of runtime objects.',
    'long_description': '# einspect\n\n<!-- start badges -->\n\n[![Build](https://github.com/ionite34/einspect/actions/workflows/build.yml/badge.svg)](https://github.com/ionite34/einspect/actions/workflows/build.yml)\n[![codecov](https://codecov.io/gh/ionite34/einspect/branch/main/graph/badge.svg?token=v71SdG5Bo6)](https://codecov.io/gh/ionite34/einspect)\n[![security](https://snyk-widget.herokuapp.com/badge/pip/einspect/badge.svg)](https://security.snyk.io/package/pip/einspect)\n\n[![PyPI](https://img.shields.io/pypi/v/einspect)][pypi]\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/einspect)][pypi]\n\n[pypi]: https://pypi.org/project/einspect/\n\n<!-- end badges -->\n\n> Extended Inspections for CPython\n\n## [Documentation](https://docs.ionite.io/einspect)\n\n- [View and modify memory structures of live objects.](#check-detailed-states-of-built-in-objects)\n- [Able to mutate immutable objects like tuples and ints.](#mutate-tuples-strings-ints-or-other-immutable-types)\n- [Modify slot functions or attributes of built-in types.](#modify-attributes-of-built-in-types-get-original-attributes-with-orig)\n- [Fully typed, extensible framework in pure Python.](#move-objects-in-memory)\n\n<!-- start intro -->\n\n## Check detailed states of built-in objects\n```python\nfrom einspect import view\n\nls = [1, 2, 3]\nv = view(ls)\nprint(v.info())\n```\n```python\nPyListObject(at 0x2833738):\n   ob_refcnt: Py_ssize_t = 5\n   ob_type: *PyTypeObject = &[list]\n   ob_item: **PyObject = &[&[1], &[2], &[3]]\n   allocated: Py_ssize_t = 4\n```\n\n[doc_tuple_view]: https://docs.ionite.io/einspect/api/views/view_tuple.html#einspect.views.view_tuple\n[doc_str_view]: https://docs.ionite.io/einspect/api/views/view_str.html#einspect.views.view_str\n[py_doc_mutable_seq]: https://docs.python.org/3/library/stdtypes.html#mutable-sequence-types\n## Mutate tuples, strings, ints, or other immutable types\n> [TupleView][doc_tuple_view] and [StrView][doc_str_view] supports all [MutableSequence][py_doc_mutable_seq] methods (append, extend, insert, pop, remove, reverse, clear).\n\n> ⚠️ A note on [safety.](#safety)\n```python\nfrom einspect import view\n\ntup = (1, 2)\nv = view(tup)\n\nv[1] = 500\nprint(tup)      # (1, 500)\nv.append(3)\nprint(tup)      # (1, 500, 3)\n\ndel v[:2]\nprint(tup)      # (3,)\nprint(v.pop())  # 3\n\nv.extend([1, 2])\nprint(tup)      # (1, 2)\n\nv.clear()\nprint(tup)      # ()\n```\n```python\nfrom einspect import view\n\ntext = "hello"\n\nv = view(text)\nv[1] = "3"\nv[4:] = "o~"\nv.append("!")\n\nprint(text)  # h3llo~!\nv.reverse()\nprint(text)  # !~oll3h\n```\n```python\nfrom einspect import view\n\nn = 500\nview(n).value = 10\n\nprint(500)        # 10\nprint(500 == 10)  # True\n```\n\n## Modify attributes of built-in types, get original attributes with `orig`\n```python\nfrom einspect import view, orig\n\nv = view(int)\nv["__name__"] = "custom_int"\nv["__iter__"] = lambda s: iter(range(s))\nv["__repr__"] = lambda s: "custom: " + orig(int).__repr__(s)\n\nprint(int)\nfor i in 3:\n    print(i)\n```\n```\n<class \'custom_int\'>\ncustom: 0\ncustom: 1\ncustom: 2\n```\n\n## Implement methods on built-in types\n> See the [Extending Types](https://docs.ionite.io/einspect/extending_types.html) docs page for more information.\n```python\nfrom einspect import impl, orig\n\n@impl(int)\ndef __add__(self, other):\n    other = int(other)\n    return orig(int).__add__(self, other)\n\nprint(50 + "25")  # 75\n```\n\n## Move objects in memory\n```python\nfrom einspect import view\n\ns = "meaning of life"\n\nv = view(s)\nwith v.unsafe():\n    v <<= 42\n\nprint("meaning of life")        # 42\nprint("meaning of life" == 42)  # True\n```\n\n<!-- end intro -->\n\n## Fully typed interface\n<img width="551" alt="image" src="https://user-images.githubusercontent.com/13956642/211129165-38a1c405-9d54-413c-962e-6917f1f3c2a1.png">\n\n## Safety\n\n This project is mainly for learning purposes or inspecting and debugging CPython internals for development and fun. You should not violate language conventions like mutability in production software and libraries.\n\nThe interpreter makes assumptions regarding types that are immutable, and changing them causes all those usages to be affected. While the intent of the project is to make a memory-correct mutation without further side effects, there can be very significant runtime implications of mutating interned strings with lots of shared references, including interpreter crashes.\n\nFor example, some strings like "abc" are interned and used by the interpreter. Changing them changes all usages of them, even attribute calls like `collections.abc`.\n\n> The spirit of safety maintained by einspect is to do with memory layouts, not functional effects.\n\n### For example, appending to tuple views (without an unsafe context) will check that the resize can fit within allocated memory\n```python\nfrom einspect import view\n\ntup = (1, 2)\nv = view(tup)\n\nv.append(3)\nprint(tup)  # (1, 2, 3)\n\nv.append(4)\n# UnsafeError: insert required tuple to be resized beyond current memory allocation. Enter an unsafe context to allow this.\n```\n- Despite this, mutating shared references like empty tuples can cause issues in interpreter shutdown and other runtime operations.\n```python\nfrom einspect import view\n\ntup = ()\nview(tup).append(1)\n```\n```\nException ignored in: <module \'threading\' from \'/lib/python3.11/threading.py\'>\nTraceback (most recent call last):\n  File "/lib/python3.11/threading.py", line 1563, in _shutdown\n    _main_thread._stop()\n  File "/lib/python3.11/threading.py", line 1067, in _stop\n    with _shutdown_locks_lock:\nTypeError: \'str\' object cannot be interpreted as an integer\n```\n\n### Similarly, memory moves are also checked for GC-header compatibility and allocation sizes\n```python\nfrom einspect import view\n\nv = view(101)\nv <<= 2\n\nprint(101)  # 2\n\nv <<= "hello"\n# UnsafeError: memory move of 54 bytes into allocated space of 32 bytes is out of bounds. Enter an unsafe context to allow this.\n```\n\n- However, this will not check the fact that small integers between (-5, 256) are interned and used by the interpreter. Changing them may cause issues in any library or interpreter Python code.\n```python\nfrom einspect import view\n\nview(0) << 100\n\nexit()\n# sys:1: ImportWarning: can\'t resolve package from __spec__ or __package__, falling back on __name__ and __path__\n# IndexError: string index out of range\n```\n\n## Table of Contents\n- [Views](#views)\n  - [Using the `einspect.view` constructor](#using-the-einspectview-constructor)\n  - [Inspecting struct attributes](#inspecting-struct-attributes)\n\n## Views\n\n### Using the `einspect.view` constructor\n\nThis is the recommended and simplest way to create a `View` onto an object. Equivalent to constructing a specific `View` subtype from `einspect.views`, except the choice of subtype is automatic based on object type.\n\n```python\nfrom einspect import view\n\nprint(view(1))\nprint(view("hello"))\nprint(view([1, 2]))\nprint(view((1, 2)))\n```\n> ```\n> IntView(<PyLongObject at 0x102058920>)\n> StrView(<PyUnicodeObject at 0x100f12ab0>)\n> ListView(<PyListObject at 0x10124f800>)\n> TupleView(<PyTupleObject at 0x100f19a00>)\n> ```\n\n### Inspecting struct attributes\n\nAttributes of the underlying C Struct of objects can be accessed through the view\'s properties.\n```python\nfrom einspect import view\n\nls = [1, 2]\nv = view(ls)\n\n# Inherited from PyObject\nprint(v.ref_count)  # ob_refcnt\nprint(v.type)       # ob_type\n# Inherited from PyVarObject\nprint(v.size)       # ob_size\n# From PyListObject\nprint(v.item)       # ob_item\nprint(v.allocated)  # allocated\n```\n> ```\n> 4\n> <class \'tuple\'>\n> 3\n> <einspect.structs.c_long_Array_3 object at 0x105038ed0>\n> ```\n\n## 2. Writing to view attributes\n\nWriting to these attributes will affect the underlying object of the view.\n\nNote that most memory-unsafe attribute modifications require entering an unsafe context manager with `View.unsafe()`\n```python\nwith v.unsafe():\n    v.size -= 1\n\nprint(obj)\n```\n> `(1, 2)`\n\nSince `items` is an array of integer pointers to python objects, they can be replaced by `id()` addresses to modify\nindex items in the tuple.\n```python\nfrom einspect import view\n\ntup = (100, 200)\n\nwith view(tup).unsafe() as v:\n    s = "dog"\n    v.item[0] = id(s)\n\nprint(tup)\n```\n> ```\n> (\'dog\', 200)\n>\n> >> Process finished with exit code 139 (interrupted by signal 11: SIGSEGV)\n> ```\n\nSo here we did set the item at index 0 with our new item, the string `"dog"`, but this also caused a segmentation fault.\nNote that the act of setting an item in containers like tuples and lists "steals" a reference to the object, even\nif we only supplied the address pointer.\n\nTo make this safe, we will have to manually increment a ref-count before the new item is assigned. To do this we can\neither create a `view` of our new item, and increment its `ref_count += 1`, or use the apis from `einspect.api`, which\nare pre-typed implementations of `ctypes.pythonapi` methods.\n```python\nfrom einspect import view\nfrom einspect.api import Py\n\ntup = (100, 200)\n\nwith view(tup).unsafe() as v:\n    a = "bird"\n    Py.IncRef(a)\n    v.item[0] = id(a)\n\n    b = "kitten"\n    Py.IncRef(b)\n    v.item[1] = id(b)\n\nprint(tup)\n```\n> `(\'bird\', \'kitten\')`\n\n🎉 No more seg-faults, and we just successfully set both items in an otherwise immutable tuple.\n\nTo make the above routine easier, you can access an abstraction by simply indexing the view.\n\n```python\nfrom einspect import view\n\ntup = ("a", "b", "c")\n\nv = view(tup)\nv[0] = 123\nv[1] = "hm"\nv[2] = "🤔"\n\nprint(tup)\n```\n> `(123, \'hm\', \'🤔\')`\n',
    'author': 'ionite34',
    'author_email': 'dev@ionite.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/ionite34/einspect',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.13',
}


setup(**setup_kwargs)



from __future__ import annotations

from ctypes import (
    Union,
    addressof,
    c_char,
    c_char_p,
    c_uint,
    c_uint8,
    c_uint16,
    c_uint32,
    c_void_p,
    c_wchar,
    c_wchar_p,
    cast,
    pythonapi,
    sizeof,
)
from enum import IntEnum
from typing import Type

from typing_extensions import Annotated

from einspect.api import Py_hash_t
from einspect.protocols import bind_api
from einspect.structs.deco import struct
from einspect.structs.py_object import Fields, PyObject
from einspect.structs.traits import IsGC
from einspect.types import Array, char_p, ptr, void_p, wchar_p


def _PyUnicode_UTF8(obj: PyASCIIObject) -> c_char_p:
    return obj.astype(PyCompactUnicodeObject).utf8


def _PyUnicode_COMPACT_DATA(obj: PyASCIIObject) -> c_void_p:
    """Return a void pointer to the raw unicode buffer."""
    if obj.ascii:
        end = obj.address + sizeof(PyASCIIObject)
        return cast(end, c_void_p)
    return cast(obj.astype(PyCompactUnicodeObject).address + 1, c_void_p)


def _PyUnicode_NONCOMPACT_DATA(obj: PyASCIIObject) -> c_void_p:
    assert not obj.compact
    data = obj.astype(PyUnicodeObject).data.any
    assert data
    return data


def PyUnicode_DATA(obj: PyASCIIObject) -> c_void_p:
    if obj.compact:
        return _PyUnicode_COMPACT_DATA(obj)
    return _PyUnicode_NONCOMPACT_DATA(obj)


def PyUnicode_IS_COMPACT_ASCII(obj: PyASCIIObject) -> bool:
    return bool(obj.compact & obj.ascii)


def _PyUnicode_HAS_UTF8_MEMORY(obj: PyASCIIObject) -> bool:
    return (
        not PyUnicode_IS_COMPACT_ASCII(obj)
        and _PyUnicode_UTF8(obj)
        and cast(_PyUnicode_UTF8(obj), c_void_p).value != PyUnicode_DATA(obj).value
    )


def _PyUnicode_UTF8_LENGTH(obj) -> int:
    return obj.astype(PyCompactUnicodeObject).utf8_length


def PyUnicode_UTF8_LENGTH(obj: PyASCIIObject) -> int:
    assert obj.ready
    if PyUnicode_IS_COMPACT_ASCII(obj):
        return obj.length
    return _PyUnicode_UTF8_LENGTH(obj)


def _PyUnicode_HAS_WSTR_MEMORY(obj: PyASCIIObject) -> bool:
    """Return True if the unicode object has an allocated wstr memory block."""
    return obj.wstr and (not obj.ready or obj.wstr != PyUnicode_DATA(obj))


def PyUnicode_WSTR_LENGTH(obj: PyASCIIObject) -> int:
    if PyUnicode_IS_COMPACT_ASCII(obj):
        return obj.length
    return obj.astype(PyCompactUnicodeObject).wstr_length


class State(IntEnum):
    """State of the string object (SSTATE constants)."""

    NOT_INTERNED = 0
    INTERNED_MORTAL = 1
    INTERNED_IMMORTAL = 2


class Kind(IntEnum):
    """
    Constants for the kind field of PyUnicodeObject.
    https://docs.python.org/3/c-api/unicode.html#c.PyUnicode_KIND
    """

    PyUnicode_WCHAR = 0
    PyUnicode_1BYTE = 1
    PyUnicode_2BYTE = 2
    PyUnicode_4BYTE = 4

    def type_info(self) -> Type[c_wchar | c_uint8 | c_uint16 | c_uint32]:
        types_map = {
            0: c_wchar,
            1: c_uint8,
            2: c_uint16,
            4: c_uint32,
        }
        return types_map[int(self)]


@struct
class LegacyUnion(Union):
    any: void_p
    latin1: ptr[c_uint8]  # Py_UCS1
    ucs2: ptr[c_uint16]  # Py_UCS2
    ucs4: ptr[c_uint32]  # Py_UCS4


@struct
class PyASCIIObject(PyObject[str, None, None], IsGC):
    """
    Defines a PyUnicodeObject Structure
    """

    length: int
    hash: Annotated[int, Py_hash_t]
    interned: Annotated[int, c_uint, 2]
    kind: Annotated[int, c_uint, 3]
    compact: Annotated[int, c_uint, 1]
    ascii: Annotated[int, c_uint, 1]
    ready: Annotated[int, c_uint, 1]
    padding: Annotated[int, c_uint, 24]
    wstr: wchar_p

    def _format_fields_(self) -> Fields:
        return {
            **super()._format_fields_(),
            "length": "Py_ssize_t",
            "hash": "Py_hash_t",
            "interned": "c_uint:2",
            "kind": "c_uint:3",
            "compact": "c_uint:1",
            "ascii": "c_uint:1",
            "ready": "c_uint:1",
            "padding": "c_uint:24",
            "wstr": ("c_wchar_p", c_wchar_p),
        }

    @property
    def mem_size(self) -> int:
        """
        Return the size of the memory allocated for the string.

        Should match `unicode_sizeof_impl`
        https://github.com/python/cpython/blob/3.11/Objects/unicodeobject.c#L14120-L14149
        """
        if self.compact and self.ascii:
            size = sizeof(PyASCIIObject) + self.length + 1
        elif self.compact:
            size = sizeof(PyCompactUnicodeObject) + (self.length + 1) * Kind(self.kind)
        else:
            # If it is a two-block object, account for base object, and
            # for character block if present.
            size = sizeof(PyUnicodeObject)
            if self.astype(PyUnicodeObject).data.any:
                size += (self.length + 1) * Kind(self.kind)

        # If the wstr pointer is present, account for it unless it is shared
        # with the data pointer. Check if the data is not shared.
        if _PyUnicode_HAS_WSTR_MEMORY(self):
            size += (PyUnicode_WSTR_LENGTH(self) + 1) * sizeof(c_wchar)
        if _PyUnicode_HAS_UTF8_MEMORY(self):
            size += PyUnicode_UTF8_LENGTH(self) + 1

        return size

    @property
    def buffer(self) -> Array:
        utf8_length_offset: int = PyUnicodeObject.utf8_length.offset  # type: ignore
        data_offset: int = PyUnicodeObject.data.offset  # type: ignore
        addr = addressof(self)

        if self.compact:
            # Get the str subtype type mapping
            subtype = Kind(self.kind).type_info()
            if self.ascii:
                # ASCII buffer comes right after wstr
                subtype = c_char
                addr += utf8_length_offset
            else:
                # UCS1/2/4 buffer comes right after wstr
                addr += data_offset
            return (subtype * self.length).from_address(addr)

        if self.kind == Kind.PyUnicode_WCHAR:
            # Note that this goes with wstr_length, not length!
            return self.wstr  # type: ignore
        elif self.kind == Kind.PyUnicode_1BYTE:
            return self.data.latin1  # type: ignore
        elif self.kind == Kind.PyUnicode_2BYTE:
            return self.data.ucs2  # type: ignore
        elif self.kind == Kind.PyUnicode_4BYTE:
            return self.data.ucs4  # type: ignore

        raise ValueError(f"Unknown kind: {self.kind}")

    @bind_api(pythonapi["PyUnicode_Substring"])
    def Substring(self, start: int, end: int) -> PyUnicodeObject:
        """
        Return a substring from index start to character index end (excluded).

        Negative indices are not supported.
        """

    @bind_api(pythonapi["PyUnicode_GetLength"])
    def GetLength(self) -> int:
        """Return the length of the string in code points."""


@struct
class PyCompactUnicodeObject(PyASCIIObject):
    """
    Defines a PyCompactUnicodeObject Structure

    Non-ASCII strings allocated through PyUnicode_New use the
    PyCompactUnicodeObject structure. state.compact is set, and the data
    immediately follow the structure.
    """

    utf8_length: int  # Number of bytes in utf8, excluding the \0
    utf8: char_p  # UTF-8 representation (null-terminated)
    wstr_length: int  # Number of characters in wstr, surrogates count as two code points

    def _format_fields_(self) -> Fields:
        return {
            **super()._format_fields_(),
            "utf8_length": "Py_ssize_t",
            "utf8": ("c_char_p", c_char_p),
            "wstr_length": "Py_ssize_t",
        }


@struct
class PyUnicodeObject(PyCompactUnicodeObject):
    """Defines a PyUnicodeObject Structure."""

    data: LegacyUnion

    def _format_fields_(self) -> Fields:
        return {
            **super()._format_fields_(),
            "utf8_length": "Py_ssize_t",
            "utf8": "c_char_p",
            "wstr_length": "Py_ssize_t",
        }



from __future__ import annotations

from types import BuiltinFunctionType
from typing import Callable, TypeVar

from einspect.structs.deco import struct
from einspect.structs.include.methodobject_h import PyMethodDef
from einspect.structs.include.object_h import vectorcallfunc
from einspect.structs.py_object import PyObject
from einspect.types import ptr

_T = TypeVar("_T", BuiltinFunctionType, Callable)


@struct
class PyCFunctionObject(PyObject[_T, None, None]):
    m_ml: ptr[PyMethodDef]  # Description of the C function to call
    m_self: ptr[PyObject]  # Passed as 'self' arg to the C func, can be NULL
    m_module: ptr[PyObject]  # The __module__ attribute, can be anything
    m_weakreflist: ptr[PyObject]  # List of weak references
    vectorcall: vectorcallfunc

    @classmethod
    def from_object(cls, obj: _T) -> PyCFunctionObject[_T]:
        return super().from_object(obj)  # type: ignore
